# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Custom remote to get the load from archives present under annex"""

__docformat__ = 'restructuredtext'

import logging
import os
import os.path as op
import shutil
from operator import itemgetter
from pathlib import Path
from urllib.parse import urlparse

from annexremote import UnsupportedRequest

from datalad.consts import ARCHIVES_SPECIAL_REMOTE
from datalad.customremotes import RemoteError
from datalad.customremotes.main import main as super_main
from datalad.distribution.dataset import Dataset
from datalad.support.annexrepo import AnnexRepo
from datalad.support.archives import ArchivesCache
from datalad.support.cache import DictCache
from datalad.support.exceptions import CapturedException
from datalad.support.locking import lock_if_check_fails
from datalad.support.network import URL
from datalad.utils import (
    ensure_bytes,
    get_dataset_root,
    getpwd,
    unique,
    unlink,
)

from .base import AnnexCustomRemote

lgr = logging.getLogger('datalad.customremotes.archive')


# ####
# Preserve from previous version
# TODO: document intention
# ####
# this one might get under Runner for better output/control
def link_file_load(src, dst, dry_run=False):
    """Just a little helper to hardlink files's load
    """
    dst_dir = op.dirname(dst)
    if not op.exists(dst_dir):
        os.makedirs(dst_dir)
    if op.lexists(dst):
        lgr.log(9, "Destination file %(dst)s exists. Removing it first",
                locals())
        # TODO: how would it interact with git/git-annex
        unlink(dst)
    lgr.log(9, "Hardlinking %(src)s under %(dst)s", locals())
    src_realpath = op.realpath(src)

    try:
        os.link(src_realpath, dst)
    except (OSError, AttributeError) as e:
        # we need to catch OSError too, because Python's own logic
        # of not providing link() where it is known to be unsupported
        # (e.g. Windows) will not cover scenarios where a particular
        # filesystem simply does not implement it on an otherwise
        # sane platform (e.g. exfat on Linux)
        lgr.warning("Linking of %s failed (%s), copying file" % (src, e))
        shutil.copyfile(src_realpath, dst)
        shutil.copystat(src_realpath, dst)
    else:
        lgr.log(2, "Hardlinking finished")


# TODO: RF functionality not specific to being a custom remote (loop etc)
#       into a separate class
class ArchiveAnnexCustomRemote(AnnexCustomRemote):
    """Special custom remote allowing to obtain files from archives

     Archives must be under annex'ed themselves.
    """
    CUSTOM_REMOTE_NAME = "archive"
    SUPPORTED_SCHEMES = (
        AnnexCustomRemote._get_custom_scheme(CUSTOM_REMOTE_NAME),)
    # Since we support only 1 scheme here
    URL_SCHEME = SUPPORTED_SCHEMES[0]
    URL_PREFIX = URL_SCHEME + ":"

    COST = 500

    def __init__(self, annex, path=None, persistent_cache=True, **kwargs):
        super().__init__(annex)

        # MIH figure out what the following is all about
        # in particular path==None
        self.repo = Dataset(get_dataset_root(Path.cwd())).repo \
            if not path \
            else AnnexRepo(path, create=False, init=False)

        self.path = self.repo.path
        # annex requests load by KEY not but URL which it originally asked
        # about.  So for a key we might get back multiple URLs and as a
        # heuristic let's use the most recently asked one

        self._last_url = None  # for heuristic to choose among multiple URLs
        self._cache = ArchivesCache(self.path, persistent=persistent_cache)
        self._contentlocations = DictCache(size_limit=100)  # TODO: config ?

    def stop(self, *args):
        """Stop communication with annex"""
        self._cache.clean()

    def get_file_url(self, archive_file=None, archive_key=None, file=None,
                     size=None):
        """Given archive (file or a key) and a file -- compose URL for access

        Examples
        --------

        dl+archive:SHA256E-s176--69...3e.tar.gz#path=1/d2/2d&size=123
            when size of file within archive was known to be 123
        dl+archive:SHA256E-s176--69...3e.tar.gz#path=1/d2/2d
            when size of file within archive was not provided

        Parameters
        ----------
        size: int, optional
          Size of the file.  If not provided, will simply be empty
        """
        assert(file is not None)
        if archive_file is not None:
            if archive_key is not None:
                raise ValueError(
                    "Provide archive_file or archive_key - not both")
            archive_key = self.repo.get_file_annexinfo(archive_file)['key']
        assert(archive_key is not None)
        attrs = dict()  # looking forward for more
        if file:
            attrs['path'] = file.lstrip('/')
        if size is not None:
            attrs['size'] = size
        return str(URL(scheme=self.URL_SCHEME,
                       path=archive_key,
                       fragment=attrs))

    @property
    def cache(self):
        return self._cache

    def _parse_url(self, url):
        """Parse url and return archive key, file within archive and
        additional attributes (such as size)"""
        url = URL(url)
        assert(url.scheme == self.URL_SCHEME)
        fdict = url.fragment_dict
        if 'path' not in fdict:
            # must be old-style key/path#size=
            assert '/' in url.path, "must be of key/path format"
            key, path = url.path.split('/', 1)
        else:
            key, path = url.path, fdict.pop('path')
        if 'size' in fdict:
            fdict['size'] = int(fdict['size'])
        return key, path, fdict

    def _gen_akey_afiles(self, key, sorted=False, unique_akeys=True):
        """Given a key, yield akey, afile pairs

        if `sorted`, then first those which have extracted version in local
        cache will be yielded

        Gets determined based on urls for datalad archives

        Made "generators all the way" as an exercise but also to delay any
        checks etc until really necessary.
        """
        # we will need all URLs anyways later on ATM, so lets list() them
        # Anyways here we have a single scheme (archive) so there is not
        # much optimization possible
        urls = list(self.gen_URLS(key))

        akey_afiles = [
            self._parse_url(url)[:2]  # skip size
            for url in urls
        ]

        if unique_akeys:
            akey_afiles = unique(akey_afiles, key=itemgetter(0))

        if not sorted:
            for pair in akey_afiles:
                yield pair
            return

        # Otherwise we will go through each one

        # multiple URLs are available so we need to figure out which one
        # would be most efficient to "deal with"
        akey_afile_paths = (
            ((akey, afile), self.get_contentlocation(
                akey,
                absolute=True, verify_exists=False
            ))
            for akey, afile in akey_afiles
        )

        # by default get_contentlocation would return empty result for a key
        # which is not available locally.  But we could still have extracted
        # archive in the cache.  So we need pretty much get first all possible
        # and then only remove those which aren't present locally.  So
        # verify_exists was added
        yielded = set()
        akey_afile_paths_ = []

        # utilize cache to check which archives might already be present in the
        # cache
        for akey_afile, akey_path in akey_afile_paths:
            if akey_path and self.cache[akey_path].is_extracted:
                yield akey_afile
                yielded.add(akey_afile)
            akey_afile_paths_.append((akey_afile, akey_path))

        # replace generators with already collected ones into a list.  The idea
        # that in many cases we don't even need to create a full list and that
        # initial single yield would be enough, thus we don't need to check
        # locations etc for every possible hit
        akey_afile_paths = akey_afile_paths_

        # if not present in the cache -- check which are present
        # locally and choose that one to use, so it would get extracted
        for akey_afile, akey_path in akey_afile_paths:
            if akey_path and op.exists(akey_path):
                yielded.add(akey_afile)
                yield akey_afile

        # So no archive is present either in the cache or originally under
        # annex XXX some kind of a heuristic I guess is to use last_url ;-)
        if self._last_url and self._last_url in urls \
                and (len(urls) == len(akey_afiles)):
            akey_afile, _ = akey_afile_paths[urls.index(self._last_url)]
            yielded.add(akey_afile)
            yield akey_afile

        for akey_afile, _ in akey_afile_paths:
            if akey_afile not in yielded:
                yield akey_afile

    def get_contentlocation(self, key, absolute=False, verify_exists=True):
        """Return (relative to top or absolute) path to the file containing the key

        This is a wrapper around AnnexRepo.get_contentlocation which provides
        caching of the result (we are asking the location for the same archive
        key often)
        """
        if key not in self._contentlocations:
            fpath = self.repo.get_contentlocation(key, batch=True)
            if fpath:  # shouldn't store empty ones
                self._contentlocations[key] = fpath
        else:
            fpath = self._contentlocations[key]
            # but verify that it exists
            if verify_exists and not op.lexists(op.join(self.path, fpath)):
                # prune from cache
                del self._contentlocations[key]
                fpath = ''

        if absolute and fpath:
            return op.join(self.path, fpath)
        else:
            return fpath

    # Protocol implementation
    def checkurl(self, url):
        # TODO:  what about those MULTI and list to be returned?
        #  should we return all filenames or keys within archive?
        #  might be way too many?
        #  only if just archive portion of url is given or the one pointing
        #  to specific file?
        lgr.debug("Current directory: %s, url: %s", os.getcwd(), url)
        akey, afile, attrs = self._parse_url(url)
        size = attrs.get('size', None)

        # But reply that present only if archive is present
        # TODO: this would throw exception if not present, so this statement is
        # kinda bogus
        akey_path = self.get_contentlocation(akey, absolute=True)
        if akey_path:
            # Extract via cache only if size is not yet known
            if size is None:
                # if for testing we want to force getting the archive extracted
                efile = self.cache[akey_path].get_extracted_filename(afile)
                efile = ensure_bytes(efile)

                if op.exists(efile):
                    size = os.stat(efile).st_size

            # so it was a good successful one -- record
            self._last_url = url

            if size is None:
                return True
            else:
                # FIXME: providing filename causes annex to not even talk to
                # ask upon drop :-/
                return [dict(size=size)]  # , basename(afile))

        else:
            # TODO: theoretically we should first check if key is available
            # from any remote to know if file is available
            return False

    def checkpresent(self, key):
        # TODO: so we need to maintain mapping from urls to keys.  Then
        # we could even store the filename within archive
        # Otherwise it is unrealistic to even require to recompute key if we
        # knew the backend etc
        # The same content could be available from multiple locations within
        # the same archive, so let's not ask it twice since here we don't care
        # about "afile"
        for akey, _ in self._gen_akey_afiles(key, unique_akeys=True):
            if self.get_contentlocation(akey) \
                    or self.repo.is_available(akey, batch=True, key=True):
                return True
        # it is unclear to MIH why this must be UNKNOWN rather than FALSE
        # but this is how I found it
        raise RemoteError('Key not present')

    def remove(self, key):
        raise UnsupportedRequest('This special remote cannot remove content')
        # # TODO: proxy query to the underlying tarball under annex that if
        # # tarball was removed (not available at all) -- report success,
        # # otherwise failure (current the only one)
        # akey, afile = self._get_akey_afile(key)
        # if False:
        #     # TODO: proxy, checking present of local tarball is not
        #     # sufficient
        #     #  not exists(self.get_key_path(key)):
        #     self.send("REMOVE-SUCCESS", akey)
        # else:
        #     self.send("REMOVE-FAILURE", akey,
        #               "Removal from file archives is not supported")

    def whereis(self, key):
        return False
        # although more logical is to report back success, it leads to imho
        # more confusing duplication. See
        # http://git-annex.branchable.com/design/external_special_remote_protocol/#comment-3f9588f6a972ae566347b6f467b53b54
        # try:
        #     key, file = self._get_akey_afile(key)
        #     self.send("WHEREIS-SUCCESS", "file %s within archive %s" % (file, key))
        # except ValueError:
        #     self.send("WHEREIS-FAILURE")

    def transfer_retrieve(self, key, file):
        akeys_tried = []
        # the same file could come from multiple files within the same archive
        # So far it doesn't make sense to "try all" of them since if one fails
        # it means the others would fail too, so it makes sense to immediately
        # prune the list so we keep only the ones from unique akeys.
        # May be whenever we support extraction directly from the tarballs
        # we should go through all and choose the one easiest to get or smth.
        for akey, afile in self._gen_akey_afiles(
                key, sorted=True, unique_akeys=True):
            if not akey:
                lgr.warning("Got an empty archive key %r for key %s. Skipping",
                            akey, key)
                continue
            akeys_tried.append(akey)
            try:
                with lock_if_check_fails(
                    check=(self.get_contentlocation, (akey,)),
                    lock_path=(
                        lambda k: op.join(self.repo.path,
                                          '.git',
                                          'datalad-archives-%s' % k),
                        (akey,)),
                    operation="annex-get"
                ) as (akey_fpath, lock):
                    if lock:
                        assert not akey_fpath
                        self._annex_get_archive_by_key(akey)
                        akey_fpath = self.get_contentlocation(akey)

                if not akey_fpath:
                    raise RuntimeError(
                        "We were reported to fetch it alright but now can't "
                        "get its location.  Check logic"
                    )

                akey_path = op.join(self.repo.path, akey_fpath)
                assert op.exists(akey_path), \
                       "Key file %s is not present" % akey_path

                # Extract that bloody file from the bloody archive
                # TODO: implement/use caching, for now a simple one
                #  actually patool doesn't support extraction of a single file
                #  https://github.com/wummel/patool/issues/20
                # so
                pwd = getpwd()
                lgr.debug(
                    "Getting file %s from %s while PWD=%s",
                    afile, akey_path, pwd)
                was_extracted = self.cache[akey_path].is_extracted
                apath = self.cache[akey_path].get_extracted_file(afile)
                link_file_load(apath, file)
                if not was_extracted and self.cache[akey_path].is_extracted:
                    self.message(
                        "%s special remote is using an extraction cache "
                        "under %s. Remove it with DataLad's 'clean' "
                        "command to save disk space." %
                        (ARCHIVES_SPECIAL_REMOTE,
                         self.cache[akey_path].path),
                        type='info',
                    )
                return
            except Exception as exc:
                ce = CapturedException(exc)
                self.message(
                    "Failed to fetch {akey} containing {key}: {msg}".format(
                        akey=akey,
                        key=key,
                        # we need to get rid of any newlines, or we might
                        # break the special remote protocol
                        msg=str(ce).replace('\n', '|')
                    ))
                continue

        raise RemoteError(
            "Failed to fetch any archive containing {key}. "
            "Tried: {akeys_tried}".format(**locals())
        )

    def claimurl(self, url):
        scheme = urlparse(url).scheme
        if scheme in self.SUPPORTED_SCHEMES:
            return True
        else:
            return False

    def _annex_get_archive_by_key(self, akey):
        # TODO: make it more stringent?
        # Command could have fail to run if key was not present locally yet
        # Thus retrieve the key using annex
        # TODO: we need to report user somehow about this happening and
        # progress on the download
        from humanize import naturalsize

        from datalad.support.annexrepo import AnnexJsonProtocol

        akey_size = self.repo.get_size_from_key(akey)
        self.message(
            "To obtain some keys we need to fetch an archive "
            "of size %s"
            % (naturalsize(akey_size) if akey_size else "unknown"),
            type='info',
        )

        try:
            self.repo._call_annex(
                ["get", "--json", "--json-progress", "--key", akey],
                protocol=AnnexJsonProtocol,
            )
        except Exception:
            self.message(f'Failed to fetch archive with key {akey}')
            raise


def main():
    """cmdline entry point"""
    super_main(
        cls=ArchiveAnnexCustomRemote,
        remote_name='datalad-archives',
        description=\
        "extract content from archives (.tar{,.gz}, .zip, etc) which are "
        "in turn managed by git-annex.  See `datalad add-archive-content` "
        "command",
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Plumbing command for running a command on each (sub)dataset"""

__docformat__ = 'restructuredtext'


import inspect
import logging
import os.path as op
import sys
from typing import Union

from argparse import REMAINDER
from itertools import chain
from tempfile import mkdtemp

from datalad.cmd import NoCapture, StdOutErrCapture
from datalad.core.local.run import normalize_command

from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.interface.common_opts import (
    contains,
    dataset_state,
    jobs_opt,
    recursion_flag,
    recursion_limit,
)
from datalad.interface.results import get_status_dict
from datalad.support.constraints import (
    EnsureBool,
    EnsureChoice,
    EnsureNone,
)
from datalad.support.exceptions import InsufficientArgumentsError
from datalad.support.parallel import (
    ProducerConsumer,
    ProducerConsumerProgressLog,
    no_parentds_in_futures,
    no_subds_in_futures,
)
from datalad.support.param import Parameter
from datalad.utils import (
    SequenceFormatter,
    chpwd as chpwd_cm,
    getpwd,
    nothing_cm,
    shortened_repr,
    swallow_outputs,
)

lgr = logging.getLogger('datalad.local.foreach_dataset')


_PYTHON_CMDS = {
    'exec': exec,
    'eval': eval
}

# Centralize definition with delayed assignment for 'auto' for each case
_SAFE_TO_CONSUME_MAP = {
    'auto': lambda: 1/0,  # must be defined based on bottomup"
    'all-subds-done': no_subds_in_futures,
    'superds-done': no_parentds_in_futures,
    'always': None
}

@build_doc
class ForEachDataset(Interface):
    r"""Run a command or Python code on the dataset and/or each of its sub-datasets.

    This command provides a convenience for the cases were no dedicated DataLad command
    is provided to operate across the hierarchy of datasets. It is very similar to
    `git submodule foreach` command with the following major differences

    - by default (unless [CMD: --subdatasets-only CMD][PY: `subdatasets_only=True` PY]) it would
      include operation on the original dataset as well,
    - subdatasets could be traversed in bottom-up order,
    - can execute commands in parallel (see `jobs` option), but would account for the order,
      e.g. in bottom-up order command is executed in super-dataset only after it is executed
      in all subdatasets.

    Additional notes:

    - for execution of "external" commands we use the environment used to execute external
      git and git-annex commands.

    *Command format*

    || REFLOW >>
    [CMD: --cmd-type external CMD][PY: cmd_type='external' PY]: A few placeholders are
    supported in the command via Python format specification:
    << REFLOW ||


    - "{pwd}" will be replaced with the full path of the current working directory.
    - "{ds}" and "{refds}" will provide instances of the dataset currently
      operated on and the reference "context" dataset which was provided via ``dataset``
      argument.
    - "{tmpdir}" will be replaced with the full path of a temporary directory.
    """
    _examples_ = [
         dict(text="Aggressively  git clean  all datasets, running 5 parallel jobs",
              code_py="foreach_dataset(['git', 'clean', '-dfx'], recursive=True, jobs=5)",
              code_cmd="datalad foreach-dataset -r -J 5 git clean -dfx"),
     ]

    _params_ = dict(
        cmd=Parameter(
            args=("cmd",),
            nargs=REMAINDER,
            metavar='COMMAND',
            doc="""command for execution. [CMD: A leading '--' can be used to
            disambiguate this command from the preceding options to DataLad.
            For --cmd-type exec or eval only a single
            command argument (Python code) is supported. CMD]
            [PY: For `cmd_type='exec'` or `cmd_type='eval'` (Python code) should
            be either a string or a list with only a single item. If 'eval', the
            actual function can be passed, which will be provided all placeholders
            as keyword arguments. PY]
            """),
        cmd_type=Parameter(
            args=("--cmd-type",),
            constraints=EnsureChoice('auto', 'external', 'exec', 'eval'),
            doc="""type of the command. `external`: to be run in a child process using dataset's runner;
            'exec': Python source code to execute using 'exec(), no value returned;
            'eval': Python source code to evaluate using 'eval()', return value is placed into 'result' field.
            'auto': If used via Python API, and `cmd` is a Python function, it will use 'eval', and
            otherwise would assume 'external'."""),
        # Following options are taken from subdatasets
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""specify the dataset to operate on.  If
            no dataset is given, an attempt is made to identify the dataset
            based on the input and/or the current working directory""",
            constraints=EnsureDataset() | EnsureNone()),
        state=dataset_state,
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
        contains=contains,
        bottomup=Parameter(
            args=("--bottomup",),
            action="store_true",
            doc="""whether to report subdatasets in bottom-up order along
            each branch in the dataset tree, and not top-down."""),
        # Possible extra options to be introduced if use-case/demand comes
        # TODO: --diff  to provide `diff` record so any arbitrary  git reset --hard etc desire could be fulfilled
        # TODO: --lower-recursion-limit aka --mindepth of find to replace subdatasets-only.
        #  or may be recursion_limit could be made more sophisticated to be able to specify range
        subdatasets_only=Parameter(
            args=("-s", "--subdatasets-only"),
            action="store_true",
            doc="""whether to exclude top level dataset.  It is implied if a non-empty
            `contains` is used"""),
        output_streams=Parameter(
            args=("--output-streams", "--o-s"),
            constraints=EnsureChoice('capture', 'pass-through', 'relpath'),
            doc="""ways to handle outputs. 'capture' and return outputs from 'cmd' in the record ('stdout',
            'stderr'); 'pass-through' to the screen (and thus absent from returned record); prefix with 'relpath'
            captured output (similar to like grep does) and write to stdout and stderr. In 'relpath', relative path
            is relative to the top of the dataset if `dataset` is specified, and if not - relative to current
            directory."""),
        chpwd=Parameter(
            args=("--chpwd",),
            constraints=EnsureChoice('ds', 'pwd'),
            doc="""'ds' will change working directory to the top of the corresponding dataset. With 'pwd'
            no change of working directory will happen.
            Note that for Python commands, due to use of threads, we do not allow chdir=ds to be used
            with jobs > 1. Hint: use 'ds' and 'refds' objects' methods to execute commands in the context
            of those datasets.
            """),
        safe_to_consume=Parameter(
            args=("--safe-to-consume",),
            constraints=EnsureChoice(*_SAFE_TO_CONSUME_MAP),
            doc="""Important only in the case of parallel (jobs greater than 1) execution.
            'all-subds-done' instructs to not consider superdataset until command finished execution
            in all subdatasets (it is the value in case of 'auto' if traversal is bottomup).
            'superds-done' instructs to not process subdatasets until command finished in the super-dataset
            (it is the value in case of 'auto' in traversal is not bottom up, which is the default). With
            'always' there is no constraint on either to execute in sub or super dataset.
        """),
        jobs=jobs_opt,
    )

    @staticmethod
    @datasetmethod(name='foreach_dataset')
    @eval_results
    def __call__(
            cmd,
            *,
            cmd_type="auto",
            dataset=None,
            state='present',
            recursive=False,
            recursion_limit=None,
            contains=None,
            bottomup=False,
            subdatasets_only=False,
            output_streams='pass-through',
            chpwd='ds',  # as the most common case/scenario
            safe_to_consume='auto',
            jobs=None
            ):
        if not cmd:
            raise InsufficientArgumentsError("No command given")

        if safe_to_consume not in _SAFE_TO_CONSUME_MAP:
            raise ValueError(f"safe_to_consume - Unknown value {safe_to_consume!r}. "
                             f"Known are: {', '.join(_SAFE_TO_CONSUME_MAP)}")
        if safe_to_consume == 'auto':
            safe_to_consume_func = no_subds_in_futures if bottomup else no_parentds_in_futures
        else:
            safe_to_consume_func = _SAFE_TO_CONSUME_MAP[safe_to_consume]

        if cmd_type == 'auto':
            cmd_type = 'eval' if _is_callable(cmd) else 'external'

        python = cmd_type in _PYTHON_CMDS

        if python:
            if _is_callable(cmd):
                if cmd_type != 'eval':
                    raise ValueError(f"Can invoke provided function only in 'eval' mode. {cmd_type!r} was provided")
            else:
                # yoh decided to avoid unnecessary complication/inhomogeneity with support
                # of multiple Python commands for now; and also allow for a single string command
                # in Python interface
                if isinstance(cmd, (list, tuple)):
                    if len(cmd) > 1:
                        raise ValueError(f"Please provide a single Python expression. Got {len(cmd)}: {cmd!r}")
                    cmd = cmd[0]

                if not isinstance(cmd, str):
                    raise ValueError(f"Please provide a single Python expression or a function. Got {cmd!r}")
        else:
            if _is_callable(cmd):
                raise ValueError(f"cmd_type={cmd_type} but a function {cmd} was provided")
            protocol = NoCapture if output_streams == 'pass-through' else StdOutErrCapture

        refds = require_dataset(
            dataset, check_installed=True, purpose='foreach-dataset execution')
        pwd = getpwd()  # Note: 'run' has some more elaborate logic for this

        #
        # Producer -- datasets to act on
        #
        subdatasets_it = refds.subdatasets(
            state=state,
            recursive=recursive, recursion_limit=recursion_limit,
            contains=contains,
            bottomup=bottomup,
            result_xfm='paths',
            result_renderer='disabled',
            return_type='generator',
        )

        if subdatasets_only or contains:
            datasets_it = subdatasets_it
        else:
            if bottomup:
                datasets_it = chain(subdatasets_it, [refds.path])
            else:
                datasets_it = chain([refds.path], subdatasets_it)

        #
        # Consumer - one for all cmd_type's
        #
        def run_cmd(dspath):
            ds = Dataset(dspath)
            status_rec = get_status_dict(
                'foreach-dataset',
                ds=ds,
                path=ds.path,
                command=cmd
            )
            if not ds.is_installed():
                yield dict(
                    status_rec,
                    status="impossible",
                    message="not installed"
                )
                return
            # For consistent environment (Python) and formatting (command) similar to `run` one
            # But for Python command we provide actual ds and refds not paths
            placeholders = dict(
                pwd=pwd,
                # pass actual instances so .format could access attributes even for external commands
                ds=ds,  # if python else ds.path,
                dspath=ds.path,  # just for consistency with `run`
                refds=refds,  # if python else refds.path,
                # Check if the command contains "tmpdir" to avoid creating an
                # unnecessary temporary directory in most but not all cases.
                # Note: different from 'run' - not wrapping match within {} and doing str
                tmpdir=mkdtemp(prefix="datalad-run-") if "tmpdir" in str(cmd) else "")
            try:
                if python:
                    if isinstance(cmd, str):
                        cmd_f, cmd_a, cmd_kw = _PYTHON_CMDS[cmd_type], (cmd, placeholders), {}
                    else:
                        assert _is_callable(cmd)
                        # all placeholders are passed as kwargs to the function
                        cmd_f, cmd_a, cmd_kw = cmd, [], placeholders

                    cm = chpwd_cm(ds.path) if chpwd == 'ds' else nothing_cm()
                    with cm:
                        if output_streams == 'pass-through':
                            res = cmd_f(*cmd_a, **cmd_kw)
                            out = {}
                        elif output_streams in ('capture', 'relpath'):
                            with swallow_outputs() as cmo:
                                res = cmd_f(*cmd_a, **cmd_kw)
                                out = {
                                    'stdout': cmo.out,
                                    'stderr': cmo.err,
                                }
                        else:
                            raise RuntimeError(output_streams)
                        if cmd_type == 'eval':
                            status_rec['result'] = res
                        else:
                            assert res is None
                else:
                    try:
                        cmd_expanded = format_command(cmd, **placeholders)
                    except KeyError as exc:
                        yield dict(
                            status_rec,
                            status='impossible',
                            message=('command has an unrecognized placeholder: %s', exc))
                        return
                    out = ds.repo._git_runner.run(
                        cmd_expanded,
                        cwd=ds.path if chpwd == 'ds' else pwd,
                        protocol=protocol)
                if output_streams in ('capture', 'relpath'):
                    status_rec.update(out)
                    # provide some feedback to user in default rendering
                    if any(out.values()):
                        status_rec['message'] = shortened_repr(out, 100)
                status_rec['status'] = 'ok'
                yield status_rec
            except Exception as exc:
                # get a better version with exception handling redoing the whole
                # status dict from scratch
                yield get_status_dict(
                    'foreach-dataset',
                    ds=ds,
                    path=ds.path,
                    command=cmd,
                    exception=exc,
                    status='error',
                    message=str(exc)
                )

        if output_streams == 'pass-through':
            pc_class = ProducerConsumer
            pc_kw = {}
        else:
            pc_class = ProducerConsumerProgressLog
            pc_kw = dict(lgr=lgr, label="foreach-dataset", unit="datasets")

        if python:
            effective_jobs = pc_class.get_effective_jobs(jobs)
            if effective_jobs > 1:
                warning = ""
                if chpwd == 'ds':
                    warning += \
                        "Execution of Python commands in parallel threads while changing directory " \
                        "is not thread-safe. "
                if output_streams in ('capture', 'relpath'):
                    warning += \
                        "Execution of Python commands in parallel while capturing output is not possible."
                if warning:
                    lgr.warning("Got jobs=%d. %s We will execute without parallelization", jobs, warning)
                    jobs = 0  # no threading even between producer/consumer

        yield from pc_class(
            producer=datasets_it,
            consumer=run_cmd,
            # probably not needed
            # It is ok to start with subdatasets since top dataset already exists
            safe_to_consume=safe_to_consume_func,
            # or vice versa
            jobs=jobs,
            **pc_kw
        )

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        from datalad.interface.utils import generic_result_renderer
        if kwargs.get('output_streams') == 'relpath':
            from datalad.log import no_progress
            with no_progress():
                ds: Union[str, Dataset] = kwargs.get('dataset')
                if ds:
                    if not isinstance(ds, Dataset):
                        ds = Dataset(ds)  # so all ///, ^ etc get treated
                    refpath = ds.path
                else:
                    refpath = getpwd()
                for k in ('stdout', 'stderr'):
                    v = res.get(k)
                    if v:
                        path = res.get('path')
                        relpath = op.relpath(path, refpath) if path else ''
                        if relpath == op.curdir:
                            relpath = ''
                        if relpath and not relpath.endswith(op.sep):
                            relpath += op.sep
                        out = getattr(sys, k)
                        for l in v.splitlines():
                            out.write(f"{relpath}{l}\n")
        else:
            generic_result_renderer(res)

# Reduced version from run
def format_command(command, **kwds):
    """Plug in placeholders in `command`.

    Parameters
    ----------
    dset : Dataset
    command : str or list

    `kwds` is passed to the `format` call.

    Returns
    -------
    formatted command (str)
    """
    command = normalize_command(command)
    sfmt = SequenceFormatter()
    return sfmt.format(command, **kwds)


def _is_callable(f):
    return inspect.isfunction(f) or inspect.isbuiltin(f)



# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test Dataset class @datasetmethod bindings without possible side effects
from needed otherwise datalad.api imports.

This one to be effective should be tested first or in isolation from other
test files
"""

from ...tests.utils_pytest import (
    assert_raises,
    with_tempfile,
)
from ..dataset import Dataset


@with_tempfile(mkdir=True)
def test_datasetmethod_bound(path=None):
    ds = Dataset(path)
    # should be automagically imported/picked up if not bound already
    assert ds.create  # simplest, intfspec only 2 entries
    assert ds.download_url  # 3 entries, with dash
    assert ds.create_sibling_github  # 3 entries, 2 dashes
    assert ds.drop  # some fancy parametrization
    assert ds.get  # some fancy parametrization
    # plugins
    assert ds.addurls
    assert ds.wtf
    # if we ask for some really not known API - kaboom
    with assert_raises(AttributeError):
        ds.kaboommethod()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Interface information about credentials

Provides minimalistic interface to deal (query, request, store) with most common
types of credentials.  To be used by Authenticators
"""

__dev_doc__ = """
Possibly useful in the future 3rd part developments

https://github.com/omab/python-social-auth
    social authentication/registration mechanism with support for several
    frameworks and auth providers
"""

import time

from ..support.exceptions import (
    AccessDeniedError,
    CapturedException,
)
from ..support.keyring_ import keyring as keyring_
from ..ui import ui
from ..utils import auto_repr
from ..support.network import iso8601_to_epoch

from datalad import cfg as dlcfg
from datalad.config import anything2bool
from datalad.distribution.dataset import Dataset
from datalad.local.gitcredential import GitCredentialInterface

from logging import getLogger
lgr = getLogger('datalad.downloaders.credentials')


@auto_repr
class Credential(object):
    """Base class for different types of credentials

    Note: While subclasses can define their own `_FIELDS`, they are actually
    assumed to have particular keys by the implementation of (some)
    authenticators. `HTTPRequestsAuthenticator` and its subclasses for example,
    assume `user` and `password` to be valid keys.
    """

    # Should be defined in a subclass as a dict of fields
    # name -> dict(attributes)
    _FIELDS = None
    _KNOWN_ATTRS = {
        'hidden',    # UI should not display the value
        'repeat',    # UI should repeat entry or not. Set to False to override default logic
        'optional',  # Not mandatory thus not requested if not set
    }

    def __init__(self, name, url=None, keyring=None, auth_url=None,
                 dataset=None):
        """
        Parameters
        ----------
        name : str
            Name of the credential, as it would be identified by in the centralized
            storage of credentials
        url : str, optional
            URL string to point users to where to seek obtaining the credentials
        keyring : a keyring
            An object providing (g|s)et_password.  If None, keyring module is used
            as is
        auth_url : str, optional
            URL string this credential is going to be used with. This context
            may be needed to query some credential systems (like git-credential).
        dataset : str, Path or Dataset
            The dataset datalad is operating on with this credential. This may
            be needed for context in order to query local configs.
        """
        self.name = name
        self.url = url
        self.set_context(auth_url=auth_url, dataset=dataset)
        self._keyring = keyring or keyring_
        self._prepare()

    def _prepare(self):
        """Additional house-keeping possibly to be performed in subclasses

        Created to avoid all the passing args/kwargs of __init__ all the time
        """
        # Basic checks
        for f, fattrs in self._FIELDS.items():
            unknown_attrs = set(fattrs).difference(self._KNOWN_ATTRS)
            if unknown_attrs:
                raise ValueError("Unknown attributes %s. Known are: %s"
                                 % (unknown_attrs, self._KNOWN_ATTRS))

    def _is_field_optional(self, f):
        return self._FIELDS[f].get('optional', False)

    @property
    def is_known(self):
        """Return True if values for all fields of the credential are known"""
        try:
            return all(
                self._is_field_optional(f) or self._get_field_value(f) is not None
                for f in self._FIELDS)
        except Exception as exc:
            ce = CapturedException(exc)
            lgr.warning("Failed to query keyring: %s", ce)
            return False

    def _get_field_value(self, field):
        return dlcfg.get('datalad.credential.{name}.{field}'.format(
            name=self.name,
            field=field.replace('_', '-')
        )) or self._keyring.get(self.name, field)

    def _ask_field_value(self, f, instructions=None):
        msg = instructions if instructions else \
            ("You need to authenticate with %r credentials." % self.name +
                  (" %s provides information on how to gain access"
                   % self.url if self.url else ''))

        # provide custom options only if set for the field
        f_props = self._FIELDS[f]
        kwargs = {}
        for p in ('hidden', 'repeat'):
            if p in f_props:
                kwargs[p] = f_props[p]
        return ui.question(
            f,
            title=msg,
            **kwargs
        )

    def _ask_and_set(self, f, instructions=None):
        v = self._ask_field_value(f, instructions=instructions)
        try:
            self.set(**{f: v})
        except Exception as e:
            lgr.error("Failed to record credential field %r: %s", f, CapturedException(e))
        return v

    def enter_new(self, instructions=None, **kwargs):
        """Enter new values for the credential fields

        Parameters
        ----------
        instructions : str, optional
          If given, the auto-generated instructions based on a login-URL are
          replaced by the given string
        **kwargs
          Any given key value pairs with non-None values are used to set the
          field `key` to the given value, without asking for user input
        """
        if kwargs:
            unknown_fields = set(kwargs).difference(self._FIELDS)
            known_fields = set(self._FIELDS).difference(kwargs)
            if unknown_fields:
                raise ValueError(
                    "Unknown to %s field(s): %s.  Known but not specified: %s"
                    % (self,
                       ', '.join(sorted(unknown_fields)),
                       ', '.join(sorted(known_fields))
                       ))
        # Use ui., request credential fields corresponding to the type
        for f in self._FIELDS:
            if kwargs.get(f, None):
                # use given value, don't ask
                self.set(**{f: kwargs[f]})
            elif not self._is_field_optional(f):
                self._ask_and_set(f, instructions=instructions)

    def __call__(self, instructions=None):
        """Obtain credentials from a keyring and if any is not known -- ask

        Parameters
        ----------
        instructions : str, optional
          If given, the auto-generated instructions based on a login-URL are
          replaced by the given string
        """
        fields = {}
        # check if we shall ask for credentials, even if some are on record
        # already (but maybe they were found to need updating)
        force_reentry = dlcfg.obtain(
            'datalad.credentials.force-ask',
            valtype=anything2bool)
        for f in self._FIELDS:
            # don't query for value if we need to get a new one
            v = None if force_reentry else self._get_field_value(f)
            if not self._is_field_optional(f):
                while v is None:  # was not known
                    v = self._ask_and_set(f, instructions=instructions)
                fields[f] = v
            elif v is not None:
                fields[f] = v
        return fields

    def set(self, **kwargs):
        """Set field(s) of the credential"""
        for f, v in kwargs.items():
            if f not in self._FIELDS:
                raise ValueError("Unknown field %s. Known are: %s"
                                 % (f, self._FIELDS.keys()))
            self._keyring.set(self.name, f, v)

    def get(self, f, default=None):
        """Get a field of the credential"""
        if f not in self._FIELDS:
            raise ValueError("Unknown field %s. Known are: %s"
                             % (f, self._FIELDS.keys()))
        try:
            return self._get_field_value(f)
        except:  # MIH: what could even happen? _keyring not a dict?
            return default

    def delete(self):
        """Deletes credential values from the keyring"""
        for f in self._FIELDS:
            self._keyring.delete(self.name, f)

    def set_context(self, auth_url=None, dataset=None):
        """Set URL/dataset context after instantiation

        ATM by design the system of providers+downloaders+credentials doesn't
        necessarily provide access to that information at instantiation time of
        `Credential` objects. Hence, allow to provide this whenever we can.

        Arguments are only applied if provided. Hence, `None` does not overwrite
        a possibly already existing attribute.

        Note
        ----
        Eventually, this is going to need a major overhaul. `Providers` etc. are
        built to be mostly unaware of their context, which is why
        `get_dataset_root()` tends to be the only way of assessing what dataset
        we are operating on. This will, however, fail to detect the correct
        dataset, if it can'T be deduced from PWD, though.

        Parameters
        ----------
        auth_url : str, optional
            URL string this credential is going to be used with. This context
            may be needed to query some credential systems (like git-credential).
        dataset : str, Path or Dataset, optional
            The dataset datalad is operating on with this credential. This may
            be needed for context in order to query local configs.
        """

        # TODO: change of context should probably not be allowed. When context
        #       is actually needed for a particular credential store, this
        #       object represents such associated creds.
        #       Allowing to switch context within the same instance leads to
        #       trouble determining when exactly a reload is needed and what is
        #       to be overwritten or not.



        if auth_url:
            self.auth_url = auth_url
        if isinstance(dataset, Dataset):
            self.ds = dataset
        else:
            self.ds = Dataset(dataset) if dataset else None


class UserPassword(Credential):
    """Simple type of a credential which consists of user/password pair"""

    _FIELDS = dict([('user', {}),
                    ('password', {'hidden': True})])

    is_expired = False  # no expiration provisioned


class Token(Credential):
    """Simple type of a credential which provides a single token"""

    _FIELDS = dict([('token', {'hidden': True, 'repeat': False})])

    is_expired = False  # no expiration provisioned


class AWS_S3(Credential):
    """Credential for AWS S3 service"""

    _FIELDS = dict([('key_id', {'repeat': False}),
                    ('secret_id', {'hidden': True, 'repeat': False}),
                    ('session', {'optional': True}),
                    ('expiration', {'optional': True}),
                   ])

    @property
    def is_expired(self):
        exp = self.get('expiration', None)
        if not exp:
            return False
        exp_epoch = iso8601_to_epoch(exp)
        # -2 to artificially shorten duration of the allotment to avoid
        # possible race conditions between us checking either it has
        # already expired before submitting a request.
        expire_in = (exp_epoch - time.time() - 2) / 3600.

        lgr.debug(
            ("Credential %s has expired %.2fh ago"
                if expire_in <= 0 else "Credential %s will expire in %.2fh")
            % (self.name, expire_in))
        return expire_in <= 0


@auto_repr
class CompositeCredential(Credential):
    """Credential which represent a sequence of Credentials where front one is exposed to user
    """

    # To be defined in sub-classes
    _CREDENTIAL_CLASSES = None
    _CREDENTIAL_ADAPTERS = None

    def _prepare(self):
        assert len(self._CREDENTIAL_CLASSES) > 1, "makes sense only if there is > 1 credential"
        assert len(self._CREDENTIAL_CLASSES) == len(self._CREDENTIAL_ADAPTERS) + 1, \
            "there should be 1 less of adapter than _CREDENTIAL_CLASSES"

        for C in self._CREDENTIAL_CLASSES:
            assert issubclass(C, Credential), "%s must be a subclass of Credential" % C

        # First credential should bear the name and url
        credentials = [self._CREDENTIAL_CLASSES[0](self.name, url=self.url, keyring=self._keyring)]
        # and we just reuse its _FIELDS for _ask_field_value etc
        self._FIELDS = credentials[0]._FIELDS
        # the rest with index suffix, but storing themselves in the same keyring
        for iC, C in enumerate(self._CREDENTIAL_CLASSES[1:]):
            credentials.append(
                C(name="%s:%d" % (self.name, iC + 1), url=None, keyring=self._keyring)
            )
        self._credentials = credentials

        super(CompositeCredential, self)._prepare()

    # Here it becomes tricky, since theoretically it is the "tail"
    # ones which might expire etc, so we wouldn't exactly know what
    # new credentials outside process wanted -- it would be silly to ask
    # right away the "entry" credentials if it is just the expiration of the
    # tail credentials
    def enter_new(self):
        # should invalidate/remove all tail credentials to avoid failing attempts to login
        self._credentials[0].enter_new()
        self.refresh()

    def refresh(self):
        """Re-establish "dependent" credentials

        E.g. if code outside was reported that it expired somehow before known expiration datetime
        """
        for c in self._credentials[1:]:
            c.delete()
        # trigger re-establishing the chain
        _ = self()
        if self.is_expired:
            raise RuntimeError("Credential %s expired right upon refresh: should have not happened")

    @property
    def is_expired(self):
        return any(c.is_expired for c in self._credentials)

    def __call__(self):
        """Obtain credentials from a keyring and if any is not known -- ask"""
        # Start from the tail until we have credentials set
        idx = len(self._credentials) - 1
        for c in self._credentials[::-1]:
            if c.is_known and not c.is_expired:
                break
            idx -= 1

        if idx < 0:
            # none was known, all the same -- start with the first one
            idx = 0

        # TODO: consider moving logic of traversal into authenticator since it is
        # the one spitting out authentication error etc
        # Theoretically we could just reuse 'fields' from adapter in the next step
        # but let's do full circle, so that if any "normalization" is done by
        # Credentials we take that into account
        for c, adapter, next_c in zip(
                self._credentials[idx:],
                self._CREDENTIAL_ADAPTERS[idx:],
                self._credentials[idx + 1:]):
            fields = c()
            next_fields = adapter(self, **fields)
            next_c.set(**next_fields)

        return self._credentials[-1]()


def _nda_adapter(composite, user=None, password=None):
    from datalad.support.third.nda_aws_token_generator import NDATokenGenerator
    from .. import cfg
    nda_auth_url = cfg.obtain('datalad.externals.nda.dbserver')
    gen = NDATokenGenerator(nda_auth_url)
    lgr.debug("Generating token for NDA user %s using %s talking to %s",
              user, gen, nda_auth_url)
    try:
        token = gen.generate_token(user, password)
    except Exception as exc:  # it is really just an "Exception"
        exc_str = str(exc).lower()
        # ATM it is "Invalid username and/or password"
        # but who knows what future would bring
        if "invalid" in exc_str and ("user" in exc_str or "password" in exc_str):
            raise AccessDeniedError(exc_str)
        raise
    # There are also session and expiration we ignore... TODO anything about it?!!!
    # we could create a derived AWS_S3 which would also store session and expiration
    # and then may be Composite could use those????
    return dict(key_id=token.access_key, secret_id=token.secret_key,
                session=token.session, expiration=token.expiration)


class NDA_S3(CompositeCredential):
    """Credential to access NDA AWS

    So for NDA we need a credential which is a composite credential.
    User provides UserPassword and then some adapter generates AWS_S3
    out of it
    """
    _CREDENTIAL_CLASSES = (UserPassword, AWS_S3)
    _CREDENTIAL_ADAPTERS = (_nda_adapter,)


def _loris_adapter(composite, user=None, password=None, **kwargs):
    from datalad.support.third.loris_token_generator import LORISTokenGenerator

    gen = LORISTokenGenerator(url=composite.url)
    token = gen.generate_token(user, password)

    return dict(token=token)


class LORIS_Token(CompositeCredential):
    _CREDENTIAL_CLASSES = (UserPassword, Token)
    _CREDENTIAL_ADAPTERS = (_loris_adapter,)

    def __init__(self, name, url=None, keyring=None):
        super(CompositeCredential, self).__init__(name, url, keyring)


class GitCredential(Credential):
    """Credential to access git-credential
    """


    _FIELDS = dict([('user', {}),
                    ('password', {'hidden': True})])

    # substitute keys used within datalad by the ones used by git-credential
    _FIELDS_GIT = {'user': 'username',
                   'password': 'password'}

    is_expired = False  # no expiration provisioned

    def __init__(self, name, url=None, keyring=None,
                 auth_url=None, dataset=None):
        super().__init__(name, url=url, keyring=keyring,
                         auth_url=auth_url, dataset=dataset)

    def _get_field_value(self, field):

        from datalad import cfg as dlcfg
        cfg = self.ds.config if self.ds else dlcfg
        cfg.reload()
        from_cfg = cfg.get('datalad.credential.{name}.{field}'.format(
            name=self.name,
            field=field.replace('_', '-')
        ))

        if from_cfg:
            # config takes precedence
            return from_cfg

        # Note:
        # In opposition to the keyring approach of other `Credential`s,
        # we don't query for single values, but for an entire "description".
        # Currently required methods have to return single values, though.
        # Hence, calls to `git credential fill` could be optimised. Not easy to
        # assess when exactly this class can know whether it's context is yet to
        # be completed, though, so that another `fill` would actually yield
        # something different than before.

        self._git_cred.fill()
        git_field = self._FIELDS_GIT[field]  # translate to git-credential terms
        if git_field in self._git_cred and self._git_cred[git_field]:
            return self._git_cred[git_field]

        # we got nothing
        return

    def set(self, **kwargs):
        for f, v in kwargs.items():
            if f not in self._FIELDS:
                raise ValueError("Unknown field %s. Known are: %s"
                                 % (f, self._FIELDS.keys()))

        mapped = {self._FIELDS_GIT[k]: v for k, v in kwargs.items()}
        self._git_cred = GitCredentialInterface(url=self.auth_url, repo=self.ds,
                                                **mapped)
        self._git_cred.approve()

    def delete(self):
        """Deletes credential"""
        self._git_cred.reject()

    def set_context(self, auth_url=None, dataset=None):
        super().set_context(auth_url, dataset)
        self._git_cred = GitCredentialInterface(url=auth_url, repo=dataset)



# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test drop command"""

import os
import os.path as op
from pathlib import Path
from unittest.mock import patch

from datalad.api import (
    Dataset,
    clone,
    create,
    drop,
)
from datalad.distributed.drop import (
    _detect_nondead_annex_at_remotes,
    _detect_unpushed_revs,
)
from datalad.support.exceptions import (
    IncompleteResultsError,
    NoDatasetFound,
)
from datalad.support.gitrepo import GitRepo
from datalad.tests.utils_pytest import (
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    OBSCURE_FILENAME,
    assert_in,
    assert_in_results,
    assert_raises,
    assert_repo_status,
    assert_result_count,
    assert_status,
    eq_,
    get_deeply_nested_structure,
    known_failure_githubci_win,
    nok_,
    ok_,
    with_tempfile,
)
from datalad.utils import chpwd


ckwa = dict(result_renderer='disabled')


@with_tempfile
@with_tempfile
def test_drop_file_content(path=None, outside_path=None):
    # see docstring for test data structure
    ds = get_deeply_nested_structure(path)
    axfile_rootds = op.join("subdir", "annexed_file.txt")
    axfile_subds = op.join("subds_modified", "subdir", "annexed_file.txt")
    gitfile = op.join("subdir", "git_file.txt")

    # refuse to operate on non-ds paths
    assert_in_results(
        ds.drop(outside_path, on_failure='ignore'),
        status='error',
        message=('path not underneath the reference dataset %s', ds)
    )
    # we only have a single copy of annexed files right now
    # check that it is not dropped by default
    with assert_raises(IncompleteResultsError) as cme:
        ds.drop(axfile_rootds)
    # The --force suggestion from git-annex-drop is translated to --reckless.
    assert_in("--reckless", str(cme.value))

    # error on non-existing paths
    non_existant_relpaths = ['funky', op.join('subds_modified', 'subfunky')]
    res = ds.drop(non_existant_relpaths, on_failure='ignore')
    # only two results, one per file
    assert_result_count(res, len(non_existant_relpaths))
    for rp in non_existant_relpaths:
        assert_in_results(
            res,
            type='file',
            status='error',
            action='drop',
            error_message='File unknown to git',
            path=str(ds.pathobj / rp),
            refds=ds.path,
        )

    # drop multiple files from different datasets
    ok_(ds.repo.file_has_content(axfile_rootds))
    res = ds.drop(
        [axfile_rootds, axfile_subds],
        reckless='availability',
        jobs=2,
        on_failure='ignore')
    nok_(ds.repo.file_has_content(axfile_rootds))
    assert_result_count(res, 2)
    for rp in [axfile_rootds, axfile_subds]:
        assert_in_results(
            res,
            type='file',
            status='ok',
            action='drop',
            path=str(ds.pathobj / rp),
            refds=ds.path,
        )

    # dropping file content for files in git
    res = ds.drop(gitfile, on_failure='ignore')
    assert_result_count(res, 1)
    assert_in_results(
        res,
        type='file',
        # why is this 'notneeded' and not 'impossible'
        # if the latter, any operation on any dataset with a
        # single file in git would fail
        status='notneeded',
        action='drop',
        message="no annex'ed content",
        path=str(ds.pathobj / gitfile),
        refds=ds.path,
    )

    # modified files, we cannot drop their content
    modfile = ds.pathobj / axfile_rootds
    modfile.unlink()
    modfile.write_text('new content')
    res = ds.drop(modfile, on_failure='ignore')
    assert_in_results(
        res,
        status='impossible',
        action='drop',
        message="cannot drop modified content, save first",
        path=str(modfile),
        refds=ds.path,
    )

    # detection of untracked content
    untrackeddir = ds.pathobj / 'subds_modified' / 'subds_lvl1_modified' / \
        f'{OBSCURE_FILENAME}_directory_untracked'
    res = ds.drop(untrackeddir, on_failure='ignore')
    assert_in_results(
        res,
        status='impossible',
        action='drop',
        message="cannot drop untracked content, save first",
        path=str(untrackeddir),
        type='directory',
        refds=ds.path,
    )

    # and lastly, recursive drop
    res = ds.drop(recursive=True, on_failure='ignore')
    # there is not much to test here (we already dropped the only
    # annexed files above). however, we should see results from the top
    # ds, and the most-bottom ds
    # subdatasets
    for p in [ds.pathobj / 'subdir' / 'file_modified',
              untrackeddir]:
        assert_in_results(res, path=str(p))


@with_tempfile
@with_tempfile
def test_drop_allkeys(origpath=None, clonepath=None):
    # create a dataset with two keys, belonging to two files,
    # in two different branches
    ds = Dataset(origpath).create()
    repo = ds.repo
    repo.call_git(['checkout', '-b', 'otherbranch'])
    (ds.pathobj / 'file1').write_text('file1')
    ds.save()
    repo.call_git(['checkout', DEFAULT_BRANCH])
    (ds.pathobj / 'file2').write_text('file2')
    ds.save()

    # confirm we have two keys
    eq_(2, repo.call_annex_records(['info'])[0]['local annex keys'])

    # do it wrong first
    assert_in_results(
        ds.drop('some', what='allkeys', on_failure='ignore'),
        status='impossible',
        type='dataset',
        action='drop',
        message=(
            'cannot drop %s, with path constraints given: %s',
            'allkeys', [ds.pathobj / 'some']),
    )
    # confirm we still have two keys
    eq_(2, repo.call_annex_records(['info'])[0]['local annex keys'])

    # clone the beast and get all keys into the clone
    dsclone = clone(ds.path, clonepath)
    dsclone.repo.call_annex(['get', '--all'])
    # confirm we have two keys in the clone
    eq_(2, dsclone.repo.call_annex_records(['info'])[0]['local annex keys'])

    # now cripple availability by dropping the "hidden" key at origin
    repo.call_annex(['drop', '--branch', 'otherbranch', '--force'])
    # confirm one key left
    eq_(1, repo.call_annex_records(['info'])[0]['local annex keys'])

    # and now drop all keys from the clone, one is redundant and can be
    # dropped, the other is not and must fail
    res = dsclone.drop(what='allkeys', jobs=2, on_failure='ignore')
    # confirm one key gone, one left
    eq_(1, dsclone.repo.call_annex_records(['info'])[0]['local annex keys'])
    assert_result_count(res, 1, action='drop', status='error', type='key')
    assert_result_count(res, 1, action='drop', status='ok', type='key')
    # now force it
    res = dsclone.drop(what='allkeys', reckless='availability',
                       on_failure='ignore')
    assert_result_count(res, 1)
    assert_result_count(res, 1, action='drop', status='ok', type='key')
    # all gone
    eq_(0, dsclone.repo.call_annex_records(['info'])[0]['local annex keys'])


@with_tempfile
@with_tempfile
@with_tempfile
def test_undead_annex_detection(gitpath=None, origpath=None, clonepath=None):
    gitds = Dataset(gitpath).create(annex=False)
    # a gitrepo can be inspected too, it might just not know anything
    eq_([], _detect_nondead_annex_at_remotes(gitds.repo, 'someid'))

    origds = Dataset(origpath).create()
    origrepo = origds.repo
    # only the local repo knows about its own annex
    eq_([None], _detect_nondead_annex_at_remotes(origrepo, origrepo.uuid))

    # works with clones
    cloneds = clone(origds, clonepath)
    clonerepo = cloneds.repo
    # the clone now know two locations, itself and origin
    eq_([None, DEFAULT_REMOTE],
        _detect_nondead_annex_at_remotes(clonerepo, origrepo.uuid))
    # just from cloning the original repo location does not learn
    # about the new annex in the clone
    eq_([], _detect_nondead_annex_at_remotes(origrepo, clonerepo.uuid))
    # it will know after a push
    cloneds.push()
    eq_([None], _detect_nondead_annex_at_remotes(origrepo, clonerepo.uuid))
    # we can declare an annex dead (here done at original location)
    origrepo.call_annex(['dead', clonerepo.uuid])
    eq_([], _detect_nondead_annex_at_remotes(origrepo, clonerepo.uuid))
    # again not automatically communicated to clones
    eq_([None, DEFAULT_REMOTE],
        _detect_nondead_annex_at_remotes(clonerepo, clonerepo.uuid))
    # but a fetch will make the death known
    clonerepo.call_git(['fetch'])
    eq_([None],
        _detect_nondead_annex_at_remotes(clonerepo, clonerepo.uuid))
    # after a local git-annex branch synchronization, it is completely
    # "gone"
    clonerepo.localsync()
    eq_([],
        _detect_nondead_annex_at_remotes(clonerepo, clonerepo.uuid))


@with_tempfile
def test_uninstall_recursive(path=None):
    ds = Dataset(path)
    assert_raises(ValueError, ds.drop)

    ds = ds.create()
    subds = ds.create('sub')

    # fail to uninstall with subdatasets present
    res = ds.drop(
        what='all', reckless='availability', on_failure='ignore')
    assert_in_results(
        res,
        action='uninstall',
        path=ds.path,
        type='dataset',
        status='error',
        message=('cannot drop dataset, subdataset(s) still present '
                 '(forgot --recursive?): %s', [subds.path]),
    )
    res = ds.drop(
        what='all', reckless='availability', recursive=True,
        on_failure='ignore')
    # both datasets gone
    assert_result_count(res, 2)
    assert_result_count(res, 2, type='dataset', status='ok')
    # the subdataset is reported first
    eq_([subds.path, ds.path],
        [r.get('path') for r in res])
    # no dataset installed anymore
    eq_(ds.is_installed(), False)
    # not even a trace
    eq_(ds.pathobj.exists(), False)


@with_tempfile
@with_tempfile
def test_unpushed_state_detection(origpath=None, clonepath=None):
    origds = Dataset(origpath).create()
    # always test in annex mode
    tester = lambda x: _detect_unpushed_revs(x, True)

    origrepo = origds.repo
    # this is still a unique repo, all payload branches are
    # unpushed
    eq_([DEFAULT_BRANCH], tester(origrepo))
    origrepo.call_git(['checkout', '-b', 'otherbranch'])
    eq_([DEFAULT_BRANCH, 'otherbranch'],
        tester(origrepo))
    # let's advance the state by one
    (origds.pathobj / 'file1').write_text('some text')
    origds.save()
    # same picture
    eq_([DEFAULT_BRANCH, 'otherbranch'],
        tester(origrepo))
    # back to original branch
    origrepo.call_git(['checkout', DEFAULT_BRANCH])

    # now lets clone
    cloneds = clone(origds, clonepath)
    clonerepo = cloneds.repo
    # right after the clone there will be no unpushed changes
    eq_([], tester(clonerepo))
    # even with more than one branch in the clone
    clonerepo.call_git(['checkout', '-t', f'{DEFAULT_REMOTE}/otherbranch'])
    eq_([], tester(clonerepo))

    # let's advance the local state now
    (cloneds.pathobj / 'file2').write_text('some other text')
    cloneds.save()
    # only the modified branch is detected
    eq_(['otherbranch'], tester(clonerepo))
    # a push will bring things into the clear
    cloneds.push(to=DEFAULT_REMOTE)
    eq_([], tester(clonerepo))

    # now detach HEAD, should work and not somehow require a branch
    cloneds.repo.call_git(['reset', '--hard', 'HEAD~1'])
    # we have this state
    eq_([], tester(clonerepo))


@with_tempfile(mkdir=True)
@with_tempfile
@with_tempfile
def test_safetynet(otherpath=None, origpath=None, clonepath=None):
    # we start with a dataset that is hosted somewhere
    origds = Dataset(origpath).create()
    # a clone is made to work on the dataset
    cloneds = clone(origds, clonepath)
    # checkout a different branch at origin to simplify testing below
    origds.repo.call_git(['checkout', '-b', 'otherbranch'])

    # an untracked file is added to simulate some work
    (cloneds.pathobj / 'file1').write_text('some text')
    # now we try to drop the entire dataset in a variety of ways
    # to check that it does not happen too quickly

    # cannot use invalid mode switch values
    assert_raises(ValueError, drop, what='bananas')
    assert_raises(ValueError, drop, reckless='rubberducky')

    # cannot simple run drop somewhere and give a path to a dataset
    # to drop
    with chpwd(otherpath):
        assert_raises(NoDatasetFound, drop, clonepath, what='all')
    ok_(cloneds.is_installed())

    # refuse to remove the CWD
    with chpwd(clonepath):
        assert_raises(RuntimeError, drop, what='all')
    ok_(cloneds.is_installed())

    assert_in_results(
        cloneds.drop(what='all', on_failure='ignore'),
        message='cannot drop untracked content, save first',
        status='impossible')
    ok_(cloneds.is_installed())

    # so let's save...
    cloneds.save()
    # - branch is progressed
    # - a new key is only available here
    res = cloneds.drop(what='all', on_failure='ignore')
    assert_in_results(res, action="uninstall", status="error")
    ok_(res[0]['message'][0].startswith(
        "to-be-dropped dataset has revisions "
        "that are not available at any known sibling"))
    ok_(cloneds.is_installed())

    # so let's push -- git only
    # we cannot use git-push directly, it would not handle
    # managed branches properly
    cloneds.push(data='nothing')

    res = cloneds.drop(what='all', on_failure='ignore')
    assert_in_results(res, action="uninstall", status="error")
    ok_(res[0]['message'][0].startswith(
        "to-be-deleted local annex not declared 'dead'"))
    # some windows test setup is not very robust, explicitly
    # include the default name "origin" in the test success
    # conditions to make this more robust
    eq_(res[0]['message'][1], [DEFAULT_REMOTE])
    ok_(cloneds.is_installed())

    # announce dead
    cloneds.repo.call_annex(['dead', 'here'])
    # but just a local declaration is not good enough
    assert_in_results(
        cloneds.drop(what='all', on_failure='ignore'),
        status='error')
    ok_(cloneds.is_installed())

    # so let's push that announcement also
    cloneds.push(data='nothing')

    res = cloneds.drop(what='all', on_failure='ignore')
    assert_in_results(res, action="drop", status="error")
    ok_(res[0]['message'].startswith(
        "unsafe\nCould only verify the existence of "
        "0 out of 1 necessary"),
        msg=f"Results were {res}")
    ok_(cloneds.is_installed())

    # so let's push all
    cloneds.push()

    # and kill the beast!
    res = cloneds.drop(what='all', on_failure='ignore')
    # only now we also drop the key!
    assert_result_count(res, 2)
    assert_in_results(
        res, action='drop', type='key', status='ok', path=cloneds.path)
    assert_in_results(
        res, action='uninstall', type='dataset', status='ok', path=cloneds.path)


@with_tempfile
def test_kill(path=None):
    # create a complicated and dirty mess
    ds = get_deeply_nested_structure(path)
    # cannot use kill without recursion enabled, because there will be no
    # checks for subdatasets, hence we cannot make the impression that
    # this would be a surgical operation
    assert_raises(ValueError, ds.drop, what='all', reckless='kill')
    # wipe it out
    res = ds.drop(what='all', reckless='kill', recursive=True)
    assert_result_count(res, 1)
    assert_in_results(
        res,
        status='ok',
        path=ds.path,
        type='dataset',
        action='uninstall',
    )
    eq_(False, ds.is_installed())
    eq_(False, ds.pathobj.exists())


@with_tempfile()
def test_refuse_to_drop_cwd(path=None):
    ds = Dataset(path).create()
    (ds.pathobj / 'deep' / 'down').mkdir(parents=True)
    for p in (ds.pathobj, ds.pathobj / 'deep', ds.pathobj / 'deep' / 'down'):
        with chpwd(str(p)):
            # will never remove PWD, or anything outside the dataset
            for target in (ds.pathobj, os.curdir, os.pardir, op.join(os.pardir, os.pardir)):
                assert_raises(RuntimeError, drop, path=target, what='all')
    sub = ds.create('sub')
    subsub = sub.create('subsub')
    for p in (sub.path, subsub.path):
        with chpwd(p):
            assert_raises(RuntimeError, drop, what='all')


@with_tempfile()
def test_careless_subdataset_uninstall(path=None):
    # nested datasets
    ds = Dataset(path).create()
    subds1 = ds.create('deep1')
    ds.create('deep2')
    eq_(sorted(ds.subdatasets(result_xfm='relpaths')), ['deep1', 'deep2'])
    assert_repo_status(ds.path)
    # now we kill the sub without the parent knowing
    subds1.drop(what='all', reckless='kill', recursive=True)
    ok_(not subds1.is_installed())
    # mountpoint exists
    ok_(subds1.pathobj.exists())
    assert_repo_status(ds.path)
    # parent still knows the sub
    eq_(sorted(ds.subdatasets(result_xfm='relpaths')), ['deep1', 'deep2'])


@with_tempfile(mkdir=True)
def test_drop_nocrash_absent_subds(path=None):
    parent = Dataset(path).create()
    parent.create('sub')
    parent.drop('sub', reckless='availability')
    assert_repo_status(parent.path)
    with chpwd(path):
        assert_status('notneeded', drop('.', recursive=True))


@with_tempfile(mkdir=True)
def test_uninstall_without_super(path=None):
    # a parent dataset with a proper subdataset, and another dataset that
    # is just placed underneath the parent, but not an actual subdataset
    parent = Dataset(path).create()
    sub = parent.create('sub')
    assert_repo_status(parent.path)
    nosub = create(op.join(parent.path, 'nosub'))
    assert_repo_status(nosub.path)
    subreport = parent.subdatasets()
    assert_result_count(subreport, 1, path=sub.path)
    assert_result_count(subreport, 0, path=nosub.path)
    # it should be possible to uninstall the proper subdataset, even without
    # explicitly calling the uninstall methods of the parent -- things should
    # be figured out by datalad
    with chpwd(parent.path):
        # check False because revisions are not pushed
        drop(sub.path, what='all', reckless='availability')
    nok_(sub.is_installed())
    # no present subdatasets anymore
    subreport = parent.subdatasets()
    assert_result_count(subreport, 1)
    assert_result_count(subreport, 1, path=sub.path, state='absent')
    assert_result_count(subreport, 0, path=nosub.path)
    # but we should fail on an attempt to uninstall the non-subdataset
    with chpwd(nosub.path):
        assert_raises(RuntimeError, drop, what='all')


@with_tempfile
def test_drop_from_git(path=None):
    ds = Dataset(path).create(annex=False)
    res = ds.drop()
    assert_in_results(res, action='drop', status='notneeded')
    (ds.pathobj / 'file').write_text('some')
    ds.save()
    assert_status('notneeded', ds.drop('file'))
    assert_status('notneeded', ds.drop(what='allkeys'))


# https://github.com/datalad/datalad/issues/6180
@with_tempfile
@with_tempfile
def test_drop_uninit_annexrepo(origpath=None, path=None):
    Dataset(origpath).create()
    # just git-clone to bypass `git annex init`
    GitRepo.clone(origpath, path)
    ds = Dataset(path)
    assert_status('ok', ds.drop(what='datasets'))
    nok_(ds.is_installed())


# https://github.com/datalad/datalad/issues/6577
@with_tempfile
def test_drop_allkeys_result_contains_annex_error_messages(path=None):
    # when calling drop with allkeys, expect git-annex error
    # message(s) to be present in the result record error_message
    ds = Dataset(path).create()
    # mock annex call always yielding error-messages in this dataset
    with patch.object(ds.repo, '_call_annex_records_items_') as mock_call:
        mock_call.return_value = iter([{
            'command': 'drop',
            'success': False,
            'error-messages': ['git-annex error message here']}])
        assert_in_results(
            ds.drop(what='allkeys', on_failure='ignore'),
            error_message='git-annex error message here',
        )


# https://github.com/datalad/datalad/issues/6948
@known_failure_githubci_win  # recent git-annex, https://github.com/datalad/datalad/issues/7197
@with_tempfile
@with_tempfile
def test_nodrop_symlinked_annex(origpath=None, clonepath=None):
    # create a dataset with a key
    ds = Dataset(origpath).create(**ckwa)
    testfile = ds.pathobj / 'file1'
    testcontent = 'precious'
    testfile.write_text(testcontent)
    ds.save(**ckwa)
    rec = ds.status(testfile, annex='availability',
                    return_type='item-or-list', **ckwa)
    eq_(testcontent, Path(rec['objloc']).read_text())

    def _droptest(_ds):
        # drop refuses to drop from a symlinked annex
        if (_ds.repo.dot_git / 'annex').is_symlink():
            assert_raises(AssertionError, _ds.drop, **ckwa)
            for what in ('all', 'allkeys', 'filecontent', 'datasets'):
                assert_raises(AssertionError, _ds.drop, what=what, **ckwa)
        # but a reckless kill works without crashing
        _ds.drop(what='all', recursive=True, reckless='kill', **ckwa)
        nok_(_ds.is_installed())
        # nothing has made the original file content vanish
        _rec = ds.status(
            testfile, annex='availability',
            return_type='item-or-list', **ckwa)
        eq_(testcontent, Path(_rec['objloc']).read_text())

    # test on a clone that does not know it has key access
    dsclone1 = clone(ds.path, clonepath, reckless='ephemeral', **ckwa)
    _droptest(dsclone1)

    # test again on a clone that does think it has a key copy
    dsclone2 = clone(ds.path, clonepath, reckless='ephemeral', **ckwa)
    if not dsclone2.repo.is_managed_branch():
        # it really should not matter, but 'origin' is set to annex.ignore=true
        # on crippledFS
        # https://github.com/datalad/datalad/issues/6960
        dsclone2.get('.')
    _droptest(dsclone2)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface for creating a publication target on GitHub
"""

__docformat__ = 'restructuredtext'


import logging
import warnings
from urllib.parse import urljoin

import requests

from datalad.distributed.create_sibling_ghlike import (
    _create_sibling,
    _GitHubLike,
)
from datalad.distribution.dataset import datasetmethod
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.support.constraints import (
    EnsureChoice,
    EnsureNone,
    EnsureStr,
)
from datalad.support.param import Parameter

lgr = logging.getLogger('datalad.distribution.create_sibling_github')


class _GitHub(_GitHubLike):
    """Customizations for the GitHub platform"""
    name = 'github'
    fullname = 'GitHub'
    response_code_unauthorized = 401
    create_org_repo_endpoint = 'orgs/{organization}/repos'
    create_user_repo_endpoint = 'user/repos'
    get_authenticated_user_endpoint = 'user'
    get_repo_info_endpoint = 'repos/{user}/{repo}'
    extra_remote_settings = {
        # first make sure that annex doesn't touch this one
        # but respect any existing config
        'annex-ignore': 'true',
        # first push should separately push active branch first
        # to overcome github issue of choosing "default" branch
        # alphabetically if its name does not match the default
        # branch for the user (or organization) which now defaults
        # to "main"
        'datalad-push-default-first': 'true'
    }

    def repo_create_response(self, r):
        """
        At present the only difference from the GHlike implementation
        is the detection of an already existing repo in a 422 response.
        """
        try:
            response = r.json()
        except Exception as e:
            lgr.debug('Cannot get JSON payload of %s [%s]' , r, e)
            response = {}
        lgr.debug('%s responded with %s %s', self.fullname, r, response)
        if r.status_code == requests.codes.created:
            return dict(
                status='ok',
                preexisted=False,
                # perform some normalization
                reponame=response.get('name'),
                private=response.get('private'),
                clone_url=response.get('clone_url'),
                ssh_url=response.get('ssh_url'),
                html_url=response.get('html_url'),
                # and also return in full
                host_response=response,
            )
        elif r.status_code == requests.codes.unprocessable and \
                any('already exist' in e.get('message', '')
                    for e in response.get('errors', [])):
            return dict(
                status='impossible',
                message='repository already exists',
                preexisted=True,
            )
        elif r.status_code == self.response_code_unauthorized:
            return dict(
                status='error',
                message=('unauthorized: %s', response.get('message')),
            )
        # make sure any error-like situation causes noise
        r.raise_for_status()
        # catch-all
        raise RuntimeError(f'Unexpected host response: {response}')

    def repo_delete_request(self, orguser, reponame):
        r = requests.delete(
            urljoin(
                self.api_url,
                self.get_repo_info_endpoint.format(
                    user=orguser,
                    repo=reponame)),
            headers=self.request_headers,
        )
        # make sure any error-like situation causes noise
        r.raise_for_status()


@build_doc
class CreateSiblingGithub(Interface):
    """Create dataset sibling on GitHub.org (or an enterprise deployment).

    GitHub is a popular commercial solution for code hosting and collaborative
    development. GitHub cannot host dataset content (but see LFS,
    http://handbook.datalad.org/r.html?LFS). However, in combination with other
    data sources and siblings, publishing a dataset to GitHub can facilitate
    distribution and exchange, while still allowing any dataset consumer to
    obtain actual data content from alternative sources.

    In order to be able to use this command, a personal access token has to be
    generated on the platform (Account->Settings->Developer Settings->Personal
    access tokens->Generate new token).

    This command can be configured with
    "datalad.create-sibling-ghlike.extra-remote-settings.NETLOC.KEY=VALUE" in
    order to add any local KEY = VALUE configuration to the created sibling in
    the local `.git/config` file. NETLOC is the domain of the Github instance to
    apply the configuration for.
    This leads to a behavior that is equivalent to calling datalad's
    ``siblings('configure', ...)``||``siblings configure`` command with the
    respective KEY-VALUE pair after creating the sibling.
    The configuration, like any other, could be set at user- or system level, so
    users do not need to add this configuration to every sibling created with
    the service at NETLOC themselves.

    .. versionchanged:: 0.16
       || REFLOW >>
       The API has been aligned with the some
       ``create_sibling_...||create-sibling-...`` commands of other GitHub-like
       services, such as GOGS, GIN, GitTea.<< REFLOW ||

    .. deprecated:: 0.16
       The ``dryrun||--dryrun`` option will be removed in a future release, use
       the renamed ``dry_run||--dry-run`` option instead.
       The ``github_login||--github-login`` option will be removed in a future
       release, use the ``credential||--credential`` option instead.
       The ``github_organization||--github-organization`` option will be
       removed in a future release, prefix the reposity name with ``<org>/``
       instead.
    """

    _examples_ = [
        dict(text="Use a new sibling on GIN as a common data source that is "
                  "auto-available when cloning from GitHub",
             code_py="""\
                 > ds = Dataset('.')

                 # the sibling on GIN will host data content
                 > ds.create_sibling_gin('myrepo', name='gin')

                 # the sibling on GitHub will be used for collaborative work
                 > ds.create_sibling_github('myrepo', name='github')

                 # register the storage of the public GIN repo as a data source
                 > ds.siblings('configure', name='gin', as_common_datasrc='gin-storage')

                 # announce its availability on github
                 > ds.push(to='github')
                 """,
             code_cmd="""\
                 % datalad create-sibling-gin myrepo -s gin

                 # the sibling on GitHub will be used for collaborative work
                 % datalad create-sibling-github myrepo -s github

                 # register the storage of the public GIN repo as a data source
                 % datalad siblings configure -s gin --as-common-datasrc gin-storage

                 # announce its availability on github
                 % datalad push --to github
                 """,
             ),
    ]

    _params_ = _GitHub.create_sibling_params
    _params_['api']._doc = """\
        URL of the GitHub instance API"""
    # special casing for deprecated mode
    _params_['existing'].constraints = EnsureChoice(
        'skip', 'error', 'reconfigure', 'replace')
    _params_['existing']._doc += """\
        DEPRECATED DANGER ZONE: With 'replace', an existing repository will be
        irreversibly removed, re-initialized, and the sibling
        (re-)configured (thus implies 'reconfigure').
        `replace` could lead to data loss! In interactive sessions a
        confirmation prompt is shown, an exception is raised in non-interactive
        sessions. The 'replace' mode will be removed in a future release."""
    # deprecated options
    _params_.update(
        github_login=Parameter(
            args=('--github-login',),
            constraints=EnsureStr() | EnsureNone(),
            metavar='TOKEN',
            doc="""Deprecated, use the credential parameter instead.
            If given must be a personal access token."""),
        github_organization=Parameter(
            args=('--github-organization',),
            constraints=EnsureStr() | EnsureNone(),
            metavar='NAME',
            doc="""Deprecated, prepend a repo name with an '<orgname>/'
            prefix instead."""),
        dryrun=Parameter(
            args=("--dryrun",),
            action="store_true",
            doc="""Deprecated. Use the renamed ``dry_run||--dry-run``
            parameter"""),
    )

    @staticmethod
    @datasetmethod(name='create_sibling_github')
    @eval_results
    def __call__(
            reponame,
            *,
            dataset=None,
            recursive=False,
            recursion_limit=None,
            name='github',
            existing='error',
            github_login=None,
            credential=None,
            github_organization=None,
            access_protocol='https',
            publish_depends=None,
            private=False,
            description=None,
            dryrun=False,
            dry_run=False,
            api='https://api.github.com'):
        if dryrun and not dry_run:
            # the old one is used, and not in agreement with the new one
            warnings.warn(
                "datalad-create-sibling-github's `dryrun` option is "
                "deprecated and will be removed in a future release, "
                "use the renamed `dry_run/--dry-run` option instead.",
                DeprecationWarning)
            dry_run = dryrun

        if api == 'https://api.github.com':
            token_info = \
                'Visit https://github.com/settings/tokens to create a token.'
        else:
            token_info = 'Log into the platform, and visit [Account->' \
                         'Settings->Developer Settings->' \
                         'Personal access tokens->Generate new token] ' \
                         'to create a new token.'
        if github_login:
            warnings.warn(
                "datalad-create-sibling-github's `github_login` option is "
                "deprecated and will be removed in a future release, "
                "use the `credential` option instead.",
                DeprecationWarning)
            from unittest.mock import patch

            # shoehorn the token into an env var to read it out using the
            # normal procedures internally
            with patch.dict(
                    'os.environ',
                    {'DATALAD_CREDENTIAL_GITHUBLOGINARG_TOKEN': github_login}):
                platform = _GitHub(
                    api, 'githubloginarg', require_token=not dry_run,
                    token_info=token_info)
        else:
            platform = _GitHub(api, credential, require_token=not dry_run,
                               token_info=token_info)

        if github_organization:
            warnings.warn(
                "datalad-create-sibling-github's `github_organization` "
                "option is deprecated and will be removed in a future "
                "release, prefix the reposity name with `<org>/` instead.",
                DeprecationWarning)
            reponame = f'{github_organization}/{reponame}'

        yield from _create_sibling(
            platform=platform,
            reponame=reponame,
            dataset=dataset,
            recursive=recursive,
            recursion_limit=recursion_limit,
            name=name,
            existing=existing,
            access_protocol=access_protocol,
            publish_depends=publish_depends,
            private=private,
            description=description,
            dry_run=dry_run,
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for cmdline.helpers"""

__docformat__ = 'restructuredtext'

from datalad.tests.utils_pytest import assert_equal

from ..helpers import _fix_datalad_ri


def test_fix_datalad_ri():
    assert_equal(_fix_datalad_ri('/'), '/')
    assert_equal(_fix_datalad_ri('/a/b'), '/a/b')
    assert_equal(_fix_datalad_ri('//'), '///')
    assert_equal(_fix_datalad_ri('///'), '///')
    assert_equal(_fix_datalad_ri('//a'), '///a')
    assert_equal(_fix_datalad_ri('///a'), '///a')
    assert_equal(_fix_datalad_ri('//a/b'), '///a/b')
    assert_equal(_fix_datalad_ri('///a/b'), '///a/b')



"""DataLad command line interface"""

# ATTN!
# The rest of the code base MUST NOT import from datalad.cli
# in order to preserve the strict separation of the CLI from the
# rest.

__docformat__ = 'restructuredtext'

import logging
lgr = logging.getLogger('datalad.cli')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for interface doc wranglers.

"""

__docformat__ = 'restructuredtext'

from datalad.interface.base import (
    alter_interface_docs_for_api,
    dedent_docstring,
)
from datalad.tests.utils_pytest import (
    assert_false,
    assert_in,
    assert_not_in,
)

demo_doc = """\
    Bla bla summary

    Generic intro blurb. Ping pong ping pong ping pong ping pong.  Ping pong ping
    pong ping pong ping pong. Ping pong ping pong ping pong ping pong. Ping pong
    ping pong ping pong ping pong. Ping pong ping pong ping pong ping pong. Ping
    pong ping pong ping pong ping pong.

    || CMDLINE >>
    || REFLOW >>
    Something for the cmdline only
    Multiline!
    << REFLOW ||
    << CMDLINE ||

    || REFLOW >>
    a
    b
    << REFLOW ||
    not
       reflowed
    || REFLOW >>
    c
    << REFLOW ||

    || PYTHON >>

    || REFLOW >>
    Some Python-only bits
    Multiline!
    << REFLOW ||

    << PYTHON ||

    And an example for in-line markup: [PY: just for Python PY] and
    the other one [CMD: just for the command line CMD]. [PY: multiline
    python-only with [ brackets [] ] PY][CMD: multiline cli-only with [ brackets
    [] ] CMD]. End of demo.

    Generic appendix. Ding dong ding dong ding dong.  Ding dong ding dong ding
    dong.  Ding dong ding dong ding dong.  Ding dong ding dong ding dong.  Ding
    dong ding dong ding dong.

"""

demo_paramdoc = """\

    Parameters
    ----------
    dataset : Dataset or None, optional
      something [PY: python only PY] in between [CMD: cmdline only CMD] appended
      Brackets can also be within and we can deal with [PY: multiline
      python-only with [ some brackets [] PY] [CMD: multiline cli-only [
      brackets included [ can we also have || ?] CMD].
      dataset is given, an attempt is made to identify the dataset based
      Dataset (e.g. a path), or value must be `None`. [Default: None]
"""

demo_argdoc = """\
    specify the dataset to perform the install operation
    on. If no dataset is given, an attempt is made to
    identify the dataset based on the current working
    directory and/or the `path` given. Constraints: Value
    must be a Dataset or a valid identifier of a Dataset
    (e.g. a path), or value must be `None`. [Default:
    None]
"""


def test_dedent():
    assert_false(dedent_docstring("one liner").endswith("\n"))


def test_alter_interface_docs_for_api():
    alt = alter_interface_docs_for_api(demo_doc)
    alt_l = alt.split('\n')
    # de-dented
    assert_false(alt_l[0].startswith(' '))
    assert_false(alt_l[-1].startswith(' '))
    assert_not_in('CMD', alt)
    assert_not_in('PY', alt)
    assert_not_in('REFLOW', alt)
    assert_in('a b', alt)
    assert_in('not\n   reflowed', alt)
    assert_in("Some Python-only bits Multiline!", alt)
    assert_in("Some Python-only bits", alt)
    assert_in("just for Python", alt)
    assert_not_in("just for the command line", alt)
    assert_not_in("multiline cli-only with [ brackets\n[] ]", alt)
    assert_in("multiline\npython-only with [ brackets [] ]", alt)

    altpd = alter_interface_docs_for_api(demo_paramdoc)
    assert_not_in("PY", altpd)
    assert_not_in("CMD", altpd)
    assert_in('python', altpd)
    assert_in('in between', altpd)
    assert_in('appended', altpd)
    assert_in("multiline\n  python-only with [ some brackets []", altpd)
    assert_not_in('cmdline', altpd)
    assert_not_in("multiline cli-only [\n  brackets included "
                  "[ can we also have || ?]", altpd)



# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Benchmarks for DataLad"""

import os
import sys
import os.path as osp

from subprocess import call

from datalad.runner import (
    Runner,
    GitRunner,
    StdOutErrCapture,
)


# Some tracking example -- may be we should track # of datasets.datalad.org
#import gc
#def track_num_objects():
#    return len(gc.get_objects())
#track_num_objects.unit = "objects"


from .common import SuprocBenchmarks

scripts_dir = osp.join(osp.dirname(__file__), 'scripts')
heavyout_cmd = "{} 1000".format(osp.join(scripts_dir, 'heavyout'))


class startup(SuprocBenchmarks):
    """
    Benchmarks for datalad commands startup
    """

    def setup(self):
        # we need to prepare/adjust PATH to point to installed datalad
        # We will base it on taking sys.executable
        python_path = osp.dirname(sys.executable)
        self.env = os.environ.copy()
        self.env['PATH'] = '%s:%s' % (python_path, self.env.get('PATH', ''))

    def time_import(self):
        call([sys.executable, "-c", "import datalad"])

    def time_import_api(self):
        call([sys.executable, "-c", "import datalad.api"])


class witlessrunner(SuprocBenchmarks):
    """Some rudimentary tests to see if there is no major slowdowns of Runner
    """

    def setup(self):
        self.runner = Runner()
        self.git_runner = GitRunner()

    def time_echo(self):
        self.runner.run(["echo"])

    def time_echo_gitrunner(self):
        self.git_runner.run(["echo"])

    def time_echo_gitrunner_fullcapture(self):
        self.git_runner.run(["echo"], protocol=StdOutErrCapture)



from __future__ import annotations

import random
import threading
import time
from threading import Thread

from datalad.tests.utils_pytest import assert_raises

from ..coreprotocols import StdOutCapture
from ..nonasyncrunner import ThreadedRunner
from ..protocol import (
    GeneratorMixIn,
    WitlessProtocol,
)
from .utils import py2cmd


class MinimalGeneratorProtocol(GeneratorMixIn, StdOutCapture):
    def __init__(self):
        StdOutCapture.__init__(self)
        GeneratorMixIn.__init__(self)


class MinimalStdOutGeneratorProtocol(GeneratorMixIn, StdOutCapture):
    def __init__(self):
        StdOutCapture.__init__(self)
        GeneratorMixIn.__init__(self)

    def pipe_data_received(self, fd, data):
        for line in data.decode().splitlines():
            self.send_result((fd, line))


def _runner_with_protocol(protocol: type[WitlessProtocol]) -> ThreadedRunner:
    return ThreadedRunner(
        cmd=py2cmd("for i in range(5): print(i)"),
        protocol_class=protocol,
        stdin=None)


def _run_on(runner: ThreadedRunner,
            iterate: bool,
            exceptions: list
            ):
    try:
        gen = runner.run()
        if iterate:
            for _ in gen:
                time.sleep(random.random())
    except Exception as e:
        exceptions.append(e.__class__)


def _get_run_on_threads(protocol: type[WitlessProtocol],
                        iterate: bool
                        ) -> tuple[Thread, Thread, list]:

    runner = _runner_with_protocol(protocol)

    args: tuple[ThreadedRunner, bool, list] = (runner, iterate, [])
    thread_1 = threading.Thread(target=_run_on, args=args)
    thread_2 = threading.Thread(target=_run_on, args=args)

    return thread_1, thread_2, args[2]


def _reentry_detection_run(protocol: type[WitlessProtocol],
                           iterate: bool
                           ) -> list:

    thread_1, thread_2, exception = _get_run_on_threads(protocol, iterate)

    thread_1.start()
    thread_2.start()

    thread_1.join()
    thread_2.join()
    return exception


def test_thread_serialization():
    # expect that two run calls on the same runner with a non-generator-protocol
    # do not create a runtime error (executions are serialized though)

    exceptions = _reentry_detection_run(StdOutCapture, True)
    assert exceptions == []


def test_reentry_detection():
    runner = _runner_with_protocol(MinimalGeneratorProtocol)
    runner.run()
    assert_raises(RuntimeError, runner.run)


def test_leave_handling():
    runner = _runner_with_protocol(MinimalStdOutGeneratorProtocol)
    all_results = [
        "".join(e[1] for e in runner.run())
        for _ in (0, 1)
    ]

    assert all_results[0] == all_results[1]


def test_thread_leave_handling():
    # expect no exception on repeated call to run of a runner with
    # generator-protocol, if the generator was exhausted before the second call

    thread_1, thread_2, exception = _get_run_on_threads(
        MinimalStdOutGeneratorProtocol,
        True
    )

    thread_1.start()
    thread_1.join()

    thread_2.start()
    thread_2.join()

    assert exception == []



# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test create publication target ssh web server action

"""

import logging
import os
import stat
import sys
from os import chmod
from os.path import (
    basename,
    exists,
)
from os.path import join as opj

import pytest

from datalad.api import (
    create_sibling,
    install,
    push,
)
from datalad.cmd import StdOutErrCapture
from datalad.cmd import WitlessRunner as Runner
from datalad.distribution.create_sibling import _RunnerAdapter
from datalad.support.annexrepo import AnnexRepo
from datalad.support.exceptions import (
    CommandError,
    InsufficientArgumentsError,
)
from datalad.support.gitrepo import GitRepo
from datalad.support.network import urlquote
from datalad.tests.utils_testdatasets import _mk_submodule_annex
from datalad.tests.utils_pytest import (
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    SkipTest,
    assert_dict_equal,
    assert_false,
    assert_in,
    assert_in_results,
    assert_no_errors_logged,
    assert_not_equal,
    assert_not_in,
    assert_raises,
    assert_repo_status,
    assert_result_count,
    assert_status,
    create_tree,
    eq_,
    get_mtimes_and_digests,
    get_ssh_port,
    known_failure_windows,
    ok_,
    ok_endswith,
    ok_file_has_content,
    ok_file_under_git,
    skip_if_on_windows,
    skip_if_root,
    skip_ssh,
    slow,
    swallow_logs,
    with_tempfile,
    with_testsui,
)
from datalad.utils import (
    Path,
    _path_,
    chpwd,
    on_windows,
)

from ..dataset import Dataset

lgr = logging.getLogger('datalad.tests')


_have_webui = None


def have_webui():
    """Helper for a delayed import test to break circular imports from
    deprecated extension, which gets its tests from the module too
    """
    global _have_webui
    if _have_webui is not None:
        return _have_webui

    try:
        import datalad_deprecated.sibling_webui
        from datalad_deprecated.tests.test_create_sibling_webui import (
            assert_publish_with_ui,
        )
        _have_webui = True
    except (ModuleNotFoundError, ImportError):
        _have_webui = False
    return _have_webui


if lgr.getEffectiveLevel() > logging.DEBUG:
    assert_create_sshwebserver = assert_no_errors_logged(create_sibling)
else:
    assert_create_sshwebserver = create_sibling


def assert_postupdate_hooks(path, installed=True, flat=False):
    """
    Verify that post-update hook was installed (or not, if installed=False)
    """
    from glob import glob
    if flat:
        # there is no top level dataset
        datasets = glob(opj(path, '*'))
    else:
        ds = Dataset(path)
        datasets = [ds.path] + ds.subdatasets(result_xfm='paths', recursive=True, state='present')
    for ds_ in datasets:
        ds_ = Dataset(ds_)
        hook_path = opj(ds_.path, '.git', 'hooks', 'post-update')
        if installed:
            ok_(os.path.exists(hook_path),
                msg="Missing %s" % hook_path)
        else:
            ok_(not os.path.exists(hook_path),
                msg="%s exists when it shouldn't" % hook_path)


@with_tempfile(mkdir=True)
def test_invalid_call(path=None):
    with chpwd(path):
        # ^ Change directory so that we don't fail with an
        # InvalidGitRepositoryError if the test is executed from a git
        # worktree.

        # needs a SSH URL
        assert_raises(InsufficientArgumentsError, create_sibling, '')
        assert_raises(ValueError, create_sibling, 'http://ignore.me')
        # needs an actual dataset
        assert_raises(
            ValueError,
            create_sibling, 'datalad-test:/tmp/somewhere', dataset='/nothere')
    # pre-configure a bogus remote
    ds = Dataset(path).create()
    ds.repo.add_remote('bogus', 'http://bogus.url.com')
    # fails to reconfigure by default with generated
    # and also when given an existing name
    for res in (ds.create_sibling('bogus:/tmp/somewhere', on_failure='ignore'),
                ds.create_sibling('datalad-test:/tmp/somewhere', name='bogus', on_failure='ignore')):
        assert_result_count(
            res, 1,
            status='error',
            message=(
                "sibling '%s' already configured (specify alternative name, or force reconfiguration via --existing",
                'bogus'))

    if not have_webui():
        # need an extension package
        assert_raises(RuntimeError, ds.create_sibling, '', ui=True)


@slow  # 26sec on travis
@skip_if_on_windows  # create_sibling incompatible with win servers
@skip_ssh
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_target_ssh_simple(origin=None, src_path=None, target_rootpath=None):
    ca = dict(result_renderer='disabled')
    test_fname = 'test-annex.dat'
    orig = Dataset(origin).create(**ca)
    (orig.pathobj / test_fname).write_text('some')
    orig.save(**ca)

    port = get_ssh_port("datalad-test")
    # prepare src
    source = install(
        src_path, source=origin,
        result_xfm='datasets', return_type='item-or-list')

    target_path = opj(target_rootpath, "basic")
    with swallow_logs(new_level=logging.ERROR) as cml:
        create_sibling(
            dataset=source,
            name="local_target",
            sshurl="ssh://datalad-test:{}".format(port),
            target_dir=target_path,
            ui=have_webui())
        assert_not_in('enableremote local_target failed', cml.out)

    target_gitrepo = GitRepo(target_path, create=False)  # raises if not a git repo
    assert_in("local_target", source.repo.get_remotes())
    # Both must be annex or git repositories
    src_is_annex = AnnexRepo.is_valid_repo(src_path)
    eq_(src_is_annex, AnnexRepo.is_valid_repo(target_path))
    # And target one should be known to have a known UUID within the source if annex
    if src_is_annex:
        lclcfg = AnnexRepo(src_path).config
        target_aversion = target_gitrepo.config['annex.version']
        # basic config in place
        eq_(lclcfg.get('remote.local_target.annex-ignore'), 'false')
        ok_(lclcfg.get('remote.local_target.annex-uuid'))

    # do it again without force, but use a different name to avoid initial checks
    # for existing remotes:
    with assert_raises(RuntimeError) as cm:
        assert_create_sshwebserver(
            dataset=source,
            name="local_target_alt",
            sshurl="ssh://datalad-test",
            target_dir=target_path)
    ok_(str(cm.value).startswith(
        "Target path %s already exists." % target_path))
    if src_is_annex:
        # Before we "talk" to it with git-annex directly -- we must prevent auto-upgrades
        # since the git-annex on "remote server" (e.g. docker container) could be outdated
        # and not support new git-annex repo version
        target_gitrepo.config.set('annex.autoupgraderepository', "false", scope='local')
        try:
            target_description = AnnexRepo(target_path, create=False).get_description()
        except CommandError as e:
            if 'is at unsupported version' not in str(e.stderr):
                raise
            # we just would skip this part of the test and avoid future similar query
            target_description = None
        else:
            assert target_gitrepo.config['annex.version'] == target_aversion
            assert_not_equal(target_description, None)
            assert_not_equal(target_description, target_path)
            # on yoh's laptop TMPDIR is under HOME, so things start to become
            # tricky since then target_path is shortened and we would need to know
            # remote $HOME.  To not over-complicate and still test, test only for
            # the basename of the target_path
            ok_endswith(target_description, basename(target_path))
    # now, with force and correct url, which is also used to determine
    # target_dir
    # Note: on windows absolute path is not url conform. But this way it's easy
    # to test, that ssh path is correctly used.
    if not on_windows:
        # add random file under target_path, to explicitly test existing=replace
        open(opj(target_path, 'random'), 'w').write('123')

        @with_testsui(responses=["yes"])
        def interactive_assert_create_sshwebserver():
            assert_create_sshwebserver(
                dataset=source,
                name="local_target",
                sshurl="ssh://datalad-test" + target_path,
                publish_by_default=DEFAULT_BRANCH,
                existing='replace',
                ui=have_webui(),
            )
        interactive_assert_create_sshwebserver()

        eq_("ssh://datalad-test" + urlquote(target_path),
            source.repo.get_remote_url("local_target"))
        ok_(source.repo.get_remote_url("local_target", push=True) is None)

        # ensure target tree actually replaced by source
        assert_false(exists(opj(target_path, 'random')))

        if src_is_annex:
            lclcfg = AnnexRepo(src_path).config
            eq_(lclcfg.get('remote.local_target.annex-ignore'), 'false')
            # valid uuid
            eq_(lclcfg.get('remote.local_target.annex-uuid').count('-'), 4)
            # should be added too, even if URL matches prior state
            eq_(lclcfg.get('remote.local_target.push'), DEFAULT_BRANCH)

        # again, by explicitly passing urls. Since we are on datalad-test, the
        # could use local path, but then it would not use "remote" git-annex
        # and thus potentially lead to incongruent result. So make URLs a bit
        # different by adding trailing /. to regular target_url
        target_url = "ssh://datalad-test" + target_path + "/."
        cpkwargs = dict(
            dataset=source,
            name="local_target",
            sshurl="ssh://datalad-test",
            target_dir=target_path,
            target_url=target_url,
            target_pushurl="ssh://datalad-test" + target_path,
            ui=have_webui(),
        )

        @with_testsui(responses=['yes'])
        def interactive_assert_create_sshwebserver():
            assert_create_sshwebserver(existing='replace', **cpkwargs)
        interactive_assert_create_sshwebserver()

        if src_is_annex and target_description:
            target_description = AnnexRepo(target_path,
                                           create=False).get_description()
            eq_(target_description, target_url)

        eq_(target_url,
            source.repo.get_remote_url("local_target"))
        eq_("ssh://datalad-test" + target_path,
            source.repo.get_remote_url("local_target", push=True))

        if have_webui():
            from datalad_deprecated.tests.test_create_sibling_webui import (
                assert_publish_with_ui,
            )
            assert_publish_with_ui(target_path)

        # now, push should work:
        push(dataset=source, to="local_target")

        # and we should be able to 'reconfigure'
        def process_digests_mtimes(digests, mtimes):
            # it should have triggered a hook, which would have created log and metadata files
            check_metadata = False
            for part in 'logs', 'metadata':
                metafiles = [k for k in digests if k.startswith(_path_('.git/datalad/%s/' % part))]
                # This is in effect ONLY if we have "compatible" datalad installed on remote
                # end. ATM we don't have easy way to guarantee that AFAIK (yoh),
                # so let's not check/enforce (TODO)
                # assert(len(metafiles) >= 1)  # we might have 2 logs if timestamps do not collide ;)
                # Let's actually do it to some degree
                if part == 'logs':
                    # always should have those:
                    assert (len(metafiles) >= 1)
                    with open(opj(target_path, metafiles[0])) as f:
                        if 'no datalad found' not in f.read():
                            check_metadata = True
                if part == 'metadata':
                    eq_(len(metafiles), bool(check_metadata))
                for f in metafiles:
                    digests.pop(f)
                    mtimes.pop(f)
            # and just pop some leftovers from annex
            # and ignore .git/logs content (gh-5298)
            for f in list(digests):
                if f.startswith('.git/annex/mergedrefs') \
                        or f.startswith('.git/logs/'):
                    digests.pop(f)
                    mtimes.pop(f)

        if not have_webui():
            # the rest of the test assumed that we have uploaded a UI
            return
        orig_digests, orig_mtimes = get_mtimes_and_digests(target_path)
        process_digests_mtimes(orig_digests, orig_mtimes)

        import time
        time.sleep(0.1)  # just so that mtimes change
        assert_create_sshwebserver(existing='reconfigure', **cpkwargs)
        digests, mtimes = get_mtimes_and_digests(target_path)
        process_digests_mtimes(digests, mtimes)

        assert_dict_equal(orig_digests, digests)  # nothing should change in terms of content

        # but some files should have been modified
        modified_files = {k for k in mtimes if orig_mtimes.get(k, 0) != mtimes.get(k, 0)}
        # collect which files were expected to be modified without incurring any changes
        ok_modified_files = {
            _path_('.git/hooks/post-update'), 'index.html',
        }
        ok_modified_files.add(_path_('.git/config'))
        ok_modified_files.update({f for f in digests if f.startswith(_path_('.git/datalad/web'))})
        # it seems that with some recent git behavior has changed a bit
        # and index might get touched
        if _path_('.git/index') in modified_files:
            ok_modified_files.add(_path_('.git/index'))
        ok_(modified_files.issuperset(ok_modified_files))


@slow  # 53.8496s
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
@with_tempfile
def check_target_ssh_recursive(use_ssh, origin, src_path, target_path):
    _mk_submodule_annex(origin, 'test-annex.dat', 'whatever')

    # prepare src
    source = install(src_path, source=origin, recursive=True)

    sub1 = Dataset(opj(src_path, "subm 1"))
    sub2 = Dataset(opj(src_path, "2"))

    for flat in False, True:
        target_path_ = target_dir_tpl = target_path + "-" + str(flat)

        if flat:
            target_dir_tpl += "/prefix%RELNAME"
            sep = '-'
        else:
            sep = os.path.sep

        if use_ssh:
            sshurl = "ssh://datalad-test" + target_path_
        else:
            sshurl = target_path_

        remote_name = 'remote-' + str(flat)
        with chpwd(source.path):
            assert_create_sshwebserver(
                name=remote_name,
                sshurl=sshurl,
                target_dir=target_dir_tpl,
                recursive=True,
                ui=have_webui())

        # raise if git repos were not created
        for suffix in [sep + 'subm 1', sep + '2', '']:
            target_dir = opj(target_path_, 'prefix' if flat else "").rstrip(os.path.sep) + suffix
            # raise if git repos were not created
            GitRepo(target_dir, create=False)

            if have_webui():
                from datalad_deprecated.tests.test_create_sibling_webui import (
                    assert_publish_with_ui,
                )
                assert_publish_with_ui(target_dir, rootds=not suffix, flat=flat)

        for repo in [source.repo, sub1.repo, sub2.repo]:
            assert_not_in("local_target", repo.get_remotes())

        # now, push should work:
        push(dataset=source, to=remote_name)

        # verify that we can create-sibling which was created later and possibly
        # first published in super-dataset as an empty directory
        sub3_name = 'subm 3-%s' % flat
        sub3 = source.create(sub3_name)
        # since is an empty value to force it to consider all changes since we published
        # already
        with chpwd(source.path):
            # as we discussed in gh-1495 we use the last-published state of the base
            # dataset as the indicator for modification detection with since='^'
            # hence we must not publish the base dataset on its own without recursion,
            # if we want to have this mechanism do its job
            #push(to=remote_name)  # no recursion
            out1 = assert_create_sshwebserver(
                name=remote_name,
                sshurl=sshurl,
                target_dir=target_dir_tpl,
                recursive=True,
                existing='skip',
                ui=have_webui(),
                since='^'
            )
            assert_postupdate_hooks(target_path_, installed=have_webui(), flat=flat)
            assert_result_count(out1, 1, status='ok', sibling_name=remote_name)

            # ensure that nothing is created since since is used.
            # Also cover deprecation for since='' support.  Takes just 60ms or so.
            # TODO: change or remove when removing since='' deprecation support
            out2 = assert_create_sshwebserver(
                name=remote_name,
                sshurl=sshurl,
                target_dir=target_dir_tpl,
                recursive=True,
                existing='skip',
                ui=have_webui(),
                since=''
            )
            assert_result_count(out2, 1, status='notneeded', sibling_name=remote_name)

        # so it was created on remote correctly and wasn't just skipped
        assert(Dataset(_path_(target_path_, ('prefix-' if flat else '') + sub3_name)).is_installed())
        push(dataset=source, to=remote_name, recursive=True, since='^') # just a smoke test


# we are explicitly testing deprecated since='' inside
@pytest.mark.filterwarnings("ignore: 'since' should point to commitish")
@slow  # 28 + 19sec on travis
def test_target_ssh_recursive():
    skip_if_on_windows()
    check_target_ssh_recursive(False)
    skip_ssh(check_target_ssh_recursive)(True)


@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
@with_tempfile
def check_target_ssh_since(use_ssh, origin, src_path, target_path):
    _mk_submodule_annex(origin, 'test-annex.dat', 'whatever')

    if use_ssh:
        sshurl = "ssh://datalad-test" + target_path
    else:
        sshurl = target_path
    # prepare src
    source = install(src_path, source=origin, recursive=True)
    eq_(len(source.subdatasets()), 2)
    # get a new subdataset and make sure it is committed in the super
    source.create('brandnew')
    eq_(len(source.subdatasets()), 3)
    assert_repo_status(source.path)

    # and now we create a sibling for the new subdataset only
    assert_create_sshwebserver(
        name='dominique_carrera',
        dataset=source,
        sshurl=sshurl,
        recursive=True,
        since='HEAD~1')
    # there is one thing in the target directory only, and that is the
    # remote repo of the newly added subdataset

    target = Dataset(target_path)
    ok_(not target.is_installed())  # since we didn't create it due to since
    eq_(['brandnew'], os.listdir(target_path))

    # now test functionality if we add a subdataset with a subdataset
    brandnew2 = source.create('brandnew2')
    brandnewsub = brandnew2.create('sub')
    brandnewsubsub = brandnewsub.create('sub')
    # and now we create a sibling for the new subdataset only
    assert_create_sshwebserver(
        name='dominique_carrera',
        dataset=source,
        sshurl=sshurl,
        recursive=True,
        existing='skip')
    # verify that it created the immediate subdataset
    ok_(Dataset(_path_(target_path, 'brandnew2')).is_installed())
    # but not the subs since they were not saved, thus even push would not operate
    # on them yet, so no reason for us to create them until subdatasets are saved
    ok_(not Dataset(_path_(target_path, 'brandnew2/sub')).is_installed())

    source.save(recursive=True)

    # and if repeated now -- will create those sub/sub
    assert_create_sshwebserver(
        name='dominique_carrera',
        dataset=source,
        sshurl=sshurl,
        recursive=True,
        existing='skip')
    # verify that it created the immediate subdataset
    ok_(Dataset(_path_(target_path, 'brandnew2/sub')).is_installed())
    ok_(Dataset(_path_(target_path, 'brandnew2/sub/sub')).is_installed())

    # now we will try with --since while creating even deeper nested one, and ensuring
    # it is created -- see https://github.com/datalad/datalad/issues/6596
    brandnewsubsub.create('sub')
    source.save(recursive=True)
    # and now we create a sibling for the new subdataset only
    assert_create_sshwebserver(
        name='dominique_carrera',
        dataset=source,
        sshurl=sshurl,
        recursive=True,
        existing='skip',
        since=f'{DEFAULT_REMOTE}/{DEFAULT_BRANCH}')
    # verify that it created the sub and sub/sub
    ok_(Dataset(_path_(target_path, 'brandnew2/sub/sub/sub')).is_installed())

    # we installed without web ui - no hooks should be created/enabled
    assert_postupdate_hooks(_path_(target_path, 'brandnew'), installed=False)


@slow  # 10sec + ? on travis
def test_target_ssh_since():
    skip_if_on_windows()
    skip_ssh(check_target_ssh_since)(True)
    check_target_ssh_since(False)


@skip_if_on_windows
@skip_if_root
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def check_failon_no_permissions(use_ssh, src_path, target_path):
    if use_ssh:
        sshurl = "ssh://datalad-test" + opj(target_path, 'ds')
    else:
        sshurl = opj(target_path, 'ds')
    ds = Dataset(src_path).create()
    # remove user write permissions from target path
    chmod(target_path, stat.S_IREAD | stat.S_IEXEC)
    assert_raises(
        CommandError,
        ds.create_sibling,
        name='noperm',
        sshurl=sshurl)
    # restore permissions
    chmod(target_path, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
    assert_create_sshwebserver(
        name='goodperm',
        dataset=ds,
        sshurl=sshurl)


def test_failon_no_permissions():
    skip_ssh(check_failon_no_permissions)(True)
    check_failon_no_permissions(False)


@with_tempfile(mkdir=True)
@with_tempfile
def check_replace_and_relative_sshpath(use_ssh, src_path, dst_path):
    # We need to come up with the path relative to our current home directory
    # https://github.com/datalad/datalad/issues/1653
    # but because we override HOME the HOME on the remote end would be
    # different even though a datalad-test. So we need to query it
    if use_ssh:
        from datalad import ssh_manager
        ssh = ssh_manager.get_connection('datalad-test')
        remote_home, err = ssh('pwd')
        remote_home = remote_home.rstrip('\n')
        dst_relpath = os.path.relpath(dst_path, remote_home)
        url = 'datalad-test:%s' % dst_relpath
        sibname = 'datalad-test'
    else:
        url = dst_path
        sibname = 'local'

    ds = Dataset(src_path).create()
    create_tree(ds.path, {'sub.dat': 'lots of data'})
    ds.save('sub.dat')
    res = ds.create_sibling(url, ui=have_webui())
    assert_in_results(res, action="create_sibling", sibling_name=sibname)
    published = ds.push(to=sibname, data='anything')
    assert_result_count(published, 1, path=opj(ds.path, 'sub.dat'))
    if have_webui():
        # verify that hook runs and there is nothing in stderr
        # since it exits with 0 exit even if there was a problem
        out = Runner(cwd=opj(dst_path, '.git')).run([_path_('hooks/post-update')],
                                                    protocol=StdOutErrCapture)
        assert_false(out['stdout'])
        assert_false(out['stderr'])

    # Verify that we could replace and publish no problem
    # https://github.com/datalad/datalad/issues/1656
    # Strangely it spits outs IncompleteResultsError exception atm... so just
    # checking that it fails somehow
    res = ds.create_sibling(url, on_failure='ignore')
    assert_status('error', res)
    assert_in('already configured', res[0]['message'][0])
    # "Settings" such as UI do not persist, so we specify it again
    # for the test below depending on it
    with assert_raises(RuntimeError):
        # but we cannot replace in non-interactive mode
        ds.create_sibling(url, existing='replace', ui=have_webui())

    # We don't have context manager like @with_testsui, so
    @with_testsui(responses=["yes"])
    def interactive_create_sibling():
        ds.create_sibling(url, existing='replace', ui=have_webui())
    interactive_create_sibling()

    published2 = ds.push(to=sibname, data='anything')
    assert_result_count(published2, 1, path=opj(ds.path, 'sub.dat'))

    # and one more test since in above test it would not puke ATM but just
    # not even try to copy since it assumes that file is already there
    create_tree(ds.path, {'sub2.dat': 'more data'})
    ds.save('sub2.dat')
    published3 = ds.push(to=sibname, data='nothing')  # we publish just git
    assert_result_count(published3, 0, path=opj(ds.path, 'sub2.dat'))

    if not have_webui():
        return

    # now publish "with" data, which should also trigger the hook!
    # https://github.com/datalad/datalad/issues/1658
    from glob import glob

    from datalad.consts import WEB_META_LOG
    logs_prior = glob(_path_(dst_path, WEB_META_LOG, '*'))
    published4 = ds.push(to=sibname, data='anything')
    assert_result_count(published4, 1, path=opj(ds.path, 'sub2.dat'))
    logs_post = glob(_path_(dst_path, WEB_META_LOG, '*'))
    eq_(len(logs_post), len(logs_prior) + 1)

    assert_postupdate_hooks(dst_path)


@slow  # 14 + 10sec on travis
def test_replace_and_relative_sshpath():
    skip_if_on_windows()
    skip_ssh(check_replace_and_relative_sshpath)(True)
    check_replace_and_relative_sshpath(False)


@with_tempfile(mkdir=True)
@with_tempfile(suffix="target")
def _test_target_ssh_inherit(standardgroup, ui, use_ssh, src_path, target_path):
    ds = Dataset(src_path).create()
    if use_ssh:
        target_url = 'datalad-test:%s' % target_path
    else:
        target_url = target_path
    remote = "magical"
    # for the test of setting a group, will just smoke test while using current
    # user's group
    ds.create_sibling(target_url, name=remote, shared='group', group=os.getgid(), ui=ui)  # not doing recursively
    if standardgroup:
        ds.repo.set_preferred_content('wanted', 'standard', remote)
        ds.repo.set_preferred_content('group', standardgroup, remote)
    ds.publish(to=remote)

    # now a month later we created a new subdataset... a few of the nested ones
    # A known hiccup happened when there
    # is also subsub ds added - we might incorrectly traverse and not prepare
    # sub first for subsub to inherit etc
    parent_ds = ds
    subdss = []
    nlevels = 2  # gets slow: 1 - 43 sec, 2 - 49 sec , 3 - 69 sec
    for levels in range(nlevels):
        subds = parent_ds.create('sub')
        create_tree(subds.path, {'sub.dat': 'lots of data'})
        parent_ds.save('sub', recursive=True)
        ok_file_under_git(subds.path, 'sub.dat', annexed=True)
        parent_ds = subds
        subdss.append(subds)

    target_subdss = [
        Dataset(opj(*([target_path] + ['sub'] * (i+1))))
        for i in range(nlevels)
    ]
    # since we do not have yet/thus have not used an option to record to publish
    # to that sibling by default (e.g. --set-upstream), if we run just ds.publish
    # -- should fail
    assert_result_count(
        ds.publish(on_failure='ignore'),
        1,
        status='impossible',
        message='No target sibling configured for default publication, please specify via --to')
    ds.publish(to=remote)  # should be ok, non recursive; BUT it (git or us?) would
                  # create an empty sub/ directory
    assert_postupdate_hooks(target_path, installed=ui)
    for target_sub in target_subdss:
        ok_(not target_sub.is_installed())  # still not there
    res = ds.publish(to=remote, recursive=True, on_failure='ignore')
    assert_result_count(res, 1 + len(subdss))
    assert_status(('error', 'notneeded'), res)
    assert_result_count(
        res, len(subdss),
        status='error',
        message=("Unknown target sibling '%s' for publication", 'magical'))

    # Finally publishing with inheritance
    ds.publish(to=remote, recursive=True, missing='inherit')
    assert_postupdate_hooks(target_path, installed=ui)

    def check_dss():
        # we added the remote and set all the
        for subds in subdss:
            eq_(subds.repo.get_preferred_content('wanted', remote), 'standard' if standardgroup else '')
            eq_(subds.repo.get_preferred_content('group', remote), standardgroup or '')

        for target_sub in target_subdss:
            ok_(target_sub.is_installed())  # it is there now
            eq_(target_sub.repo.config.get('core.sharedrepository'), '1')
            # and we have transferred the content
            if standardgroup and standardgroup == 'backup':
                # only then content should be copied
                ok_file_has_content(opj(target_sub.path, 'sub.dat'), 'lots of data')
            else:
                # otherwise nothing is copied by default
                assert_false(target_sub.repo.file_has_content('sub.dat'))

    check_dss()
    # and it should be ok to reconfigure the full hierarchy of datasets
    # while "inheriting". No URL must be specified, and we must not blow
    # but just issue a warning for the top level dataset which has no super,
    # so cannot inherit anything - use case is to fixup/establish the full
    # hierarchy on the remote site
    ds.save(recursive=True)  # so we have committed hierarchy for create_sibling
    with swallow_logs(logging.WARNING) as cml:
        out = ds.create_sibling(
            None, name=remote, existing="reconfigure", inherit=True,
            ui=ui, recursive=True)
        eq_(len(out), 1 + len(subdss))
        assert_in("Cannot determine super dataset", cml.out)

    check_dss()


@slow  # 49 sec
def test_target_ssh_inherit():
    skip_if_on_windows()  # create_sibling incompatible with win servers
    try:
        from datalad_deprecated.publish import Publish
    except ImportError:
        raise SkipTest('Test requires `publish()` from datalad-deprecated')
    # TODO: was waiting for resolution on
    #   https://github.com/datalad/datalad/issues/1274
    # which is now closed but this one is failing ATM, thus leaving as TODO
    # _test_target_ssh_inherit(None)      # no wanted etc
    # Takes too long so one will do with UI and another one without
    skip_ssh(_test_target_ssh_inherit)('manual', have_webui(), True)  # manual -- no load should be annex copied
    _test_target_ssh_inherit('backup', False, False)  # backup -- all data files


@with_testsui(responses=["no", "yes"])
@with_tempfile(mkdir=True)
def check_exists_interactive(use_ssh, path):
    origin = Dataset(opj(path, "origin")).create()
    sibling_path = opj(path, "sibling")

    # Initiate sibling directory with "stuff"
    create_tree(sibling_path, {'stuff': ''})

    if use_ssh:
        sshurl = 'datalad-test:' + sibling_path
    else:
        sshurl = sibling_path

    # Should fail
    with assert_raises(RuntimeError):
        origin.create_sibling(sshurl)

    # Since first response is "no" - we should fail here again:
    with assert_raises(RuntimeError):
        origin.create_sibling(sshurl, existing='replace')
    # and there should be no initiated repository
    assert not Dataset(sibling_path).is_installed()
    # But we would succeed on the 2nd try, since answer will be yes
    origin.create_sibling(sshurl, existing='replace')
    assert Dataset(sibling_path).is_installed()
    # And with_testsui should not fail with "Unused responses left"


def test_check_exists_interactive():
    skip_if_on_windows()
    skip_ssh(check_exists_interactive)(True)
    check_exists_interactive(False)


@skip_if_on_windows
@with_tempfile(mkdir=True)
def test_local_relpath(path=None):
    path = Path(path)
    ds_main = Dataset(path / "main").create()
    ds_main.create("subds")

    ds_main.create_sibling(
        name="relpath-bound", recursive=True,
        sshurl=os.path.relpath(str(path / "a"), ds_main.path))
    ok_((path / "a" / "subds").exists())

    with chpwd(path):
        create_sibling(
            dataset=ds_main.path,
            name="relpath-unbound", recursive=True,
            sshurl="b")
    ok_((path / "b" / "subds").exists())

    with chpwd(ds_main.path):
        create_sibling(
            name="relpath-unbound-dsnone", recursive=True,
            sshurl=os.path.relpath(str(path / "c"), ds_main.path))
    ok_((path / "c" / "subds").exists())


@skip_if_on_windows
@with_tempfile(mkdir=True)
def test_local_path_target_dir(path=None):
    path = Path(path)
    ds_main = Dataset(path / "main").create()

    ds_main.create_sibling(
        name="abspath-targetdir",
        sshurl=str(path / "a"), target_dir="tdir")
    ok_((path / "a" / "tdir").exists())

    ds_main.create_sibling(
        name="relpath-bound-targetdir",
        sshurl=os.path.relpath(str(path / "b"), ds_main.path),
        target_dir="tdir")
    ok_((path / "b" / "tdir").exists())

    with chpwd(path):
        create_sibling(
            dataset=ds_main.path,
            name="relpath-unbound-targetdir",
            sshurl="c", target_dir="tdir")
    ok_((path / "c" / "tdir").exists())

    ds_main.create("subds")

    ds_main.create_sibling(
        name="rec-plain-targetdir", recursive=True,
        sshurl=str(path / "d"), target_dir="tdir")
    ok_((path / "d" / "tdir" / "subds").exists())

    ds_main.create_sibling(
        name="rec-template-targetdir", recursive=True,
        sshurl=str(path / "e"), target_dir="d%RELNAME")
    ok_((path / "e" / "d").exists())
    ok_((path / "e" / "d-subds").exists())


@slow  # 12sec on Yarik's laptop
@skip_if_on_windows  # create_sibling incompatible with win servers
@skip_ssh
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_non_master_branch(src_path=None, target_path=None):
    src_path = Path(src_path)
    target_path = Path(target_path)

    ds_a = Dataset(src_path).create()
    # Rename rather than checking out another branch so that the default branch
    # doesn't exist in any state.
    ds_a.repo.call_git(["branch", "-m", DEFAULT_BRANCH, "other"])
    (ds_a.pathobj / "afile").write_text("content")
    sa = ds_a.create("sub-a")
    sa.repo.checkout("other-sub", ["-b"])
    ds_a.create("sub-b")

    ds_a.save()
    ds_a.create_sibling(
        name="sib", recursive=True,
        sshurl="ssh://datalad-test" + str(target_path / "b"))
    ds_a.push(to="sib", data="anything")

    ds_b = Dataset(target_path / "b")

    def get_branch(repo):
        return repo.get_corresponding_branch() or repo.get_active_branch()

    # The HEAD for the create-sibling matches what the branch was in
    # the original repo.
    eq_(get_branch(ds_b.repo), "other")
    ok_((ds_b.pathobj / "afile").exists())

    eq_(get_branch(Dataset(target_path / "b" / "sub-a").repo),
        "other-sub")
    eq_(get_branch(Dataset(target_path / "b" / "sub-b").repo),
        DEFAULT_BRANCH)


@known_failure_windows  # https://github.com/datalad/datalad/issues/5287
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_preserve_attrs(src=None, dest=None):
    create_tree(src, {"src": {"foo": {"bar": "This is test text."}}})
    os.utime(opj(src, "src", "foo", "bar"), (1234567890, 1234567890))
    _RunnerAdapter().put(opj(src, "src"), dest, recursive=True, preserve_attrs=True)
    s = os.stat(opj(dest, "src", "foo", "bar"))
    assert s.st_atime == 1234567890
    assert s.st_mtime == 1234567890
    with open(opj(dest, "src", "foo", "bar")) as fp:
        assert fp.read() == "This is test text."


@with_tempfile(mkdir=True)
def test_only_one_level_without_recursion(path=None):
    # this tests for https://github.com/datalad/datalad/issues/5614: accidental
    # recursion of one level by default
    path = Path(path)
    ds_main = Dataset(path / "main").create()
    ds_main.create('sub1')

    ds_main.create_sibling(
        name="dummy",
        sshurl=str(path / "toplevelsibling"))
    # this should exist
    ok_((path / 'toplevelsibling').exists())
    # this shouldn't
    assert_false(Path(path / 'toplevelsibling' / 'sub1').exists())



from __future__ import annotations

import logging
import os
import threading
from abc import (
    abstractmethod,
    ABCMeta,
)
from enum import Enum
from queue import (
    Full,
    Queue,
)
from subprocess import Popen
from typing import (
    Any,
    IO,
    Optional,
)
from datalad.utils import COPY_BUFSIZE


lgr = logging.getLogger("datalad.runner.runnerthreads")


def _try_close(file_object: Optional[IO]):
    if file_object is not None:
        try:
            file_object.close()
        except OSError:
            pass


class IOState(Enum):
    ok = "ok"
    process_exit = "process_exit"


class SignalingThread(threading.Thread):
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue]
                 ):

        super().__init__(daemon=True)
        self.identifier = identifier
        self.signal_queues = signal_queues

    def __repr__(self):
        return f"Thread<{self.identifier}>"

    def __str__(self):
        return self.__repr__()

    def signal(self,
               content: Any
               ) -> bool:
        error_occurred = False
        for signal_queue in self.signal_queues:
            try:
                signal_queue.put(content, block=True, timeout=.1)
            except Full:
                lgr.debug("timeout while trying to signal: %s", content)
                error_occurred = True
        return not error_occurred


class WaitThread(SignalingThread):
    """
    Instances of this thread wait for a process to exit and enqueue
    an exit event in the signal queues.
    """
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue],
                 process: Popen
                 ):
        super().__init__(identifier, signal_queues)
        self.process = process

    def run(self) -> None:

        lgr.log(5, "%s (%s) started", self.identifier, self)

        self.process.wait()
        self.signal((self.identifier, IOState.process_exit, None))

        lgr.log(5, "%s (%s) exiting", self.identifier, self)


class ExitingThread(SignalingThread):
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue]
                 ):

        super().__init__(identifier, signal_queues)
        self.exit_requested = False

    def request_exit(self):
        """
        Request the thread to exit. This is not guaranteed to
        have any effect, because the instance has to check for
        self.exit_requested and act accordingly. It might not
        do that.
        """
        self.exit_requested = True


class TransportThread(ExitingThread, metaclass=ABCMeta):
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue],
                 user_info: Any
                 ):

        super().__init__(identifier, signal_queues)
        self.user_info = user_info

    def __repr__(self) -> str:
        return f"Thread<({self.identifier}, {self.user_info})>"

    def __str__(self) -> str:
        return self.__repr__()

    def signal_event(self,
                     state: IOState,
                     data: Optional[bytes]
                     ) -> bool:
        return self.signal((self.user_info, state, data))

    @abstractmethod
    def read(self) -> Optional[bytes]:
        """
        Read data from source return None, if source is close,
        or destination close is required.
        """
        raise NotImplementedError

    @abstractmethod
    def write(self,
              data: bytes
              ) -> bool:
        """
        Write given data to destination, return True if data is
        written successfully, False otherwise.
        """
        raise NotImplementedError

    def run(self) -> None:

        lgr.log(5, "%s (%s) started", self.identifier, self)

        # Copy data from source queue to destination queue
        # until exit is requested. If timeouts arise, signal
        # them to the receiver via the signal queue.
        data: Optional[bytes] = b""
        while not self.exit_requested:

            data = self.read()
            # If the source sends None-data it wants
            # us to exit the thread. Signal this to
            # the downstream queues (which might or might
            # not be contain the output queue),
            # and exit the thread.
            if data is None:
                break

            if self.exit_requested:
                break

            succeeded = self.write(data)
            if not succeeded:
                break

        self.signal_event(IOState.ok, None)
        lgr.log(
            5,
            "%s (%s) exiting (exit_requested: %s, last data: %s)",
            self.identifier,
            self,
            self.exit_requested, data)


class ReadThread(TransportThread):
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue],
                 user_info: Any,
                 source: IO,
                 destination_queue: Queue,
                 length: int = COPY_BUFSIZE
                 ):

        super().__init__(identifier, signal_queues, user_info)
        self.source = source
        self.destination_queue = destination_queue
        self.length = length

    def read(self) -> Optional[bytes]:
        try:
            data = os.read(self.source.fileno(), self.length)
        except (ValueError, OSError):
            # The destination was most likely closed, nevertheless,
            # try to close it and indicate EOF.
            _try_close(self.source)
            return None
        return data or None

    def write(self,
              data: bytes
              ) -> bool:

        # We write to an unlimited queue, no need for timeout checking.
        self.destination_queue.put((self.user_info, IOState.ok, data))
        return True


class WriteThread(TransportThread):
    def __init__(self,
                 identifier: str,
                 signal_queues: list[Queue],
                 user_info: Any,
                 source_queue: Queue,
                 destination: IO
                 ):

        super().__init__(identifier, signal_queues, user_info)
        self.source_queue = source_queue
        self.destination = destination

    def read(self) -> Optional[bytes]:
        data = self.source_queue.get()
        if data is None:
            # Close stdin file descriptor here, since we know that no more
            # data will be sent to stdin.
            _try_close(self.destination)
        return data

    def write(self,
              data: bytes,
              ) -> bool:
        try:
            written = 0
            while written < len(data):
                written += os.write(
                    self.destination.fileno(),
                    data[written:])
                if self.exit_requested:
                    return written == len(data)
        except (BrokenPipeError, OSError, ValueError):
            # The destination was most likely closed, nevertheless,
            # try to close it and indicate EOF.
            _try_close(self.destination)
            return False
        return True



from __future__ import annotations

from datalad.tests.utils_pytest import assert_equal

from ..exception import (
    CommandError,
    _format_json_error_messages,
)


def get_json_objects(object_count, message_count) -> list[dict]:
    return [
        {
            "success": index % 2 == 0,
            "file": f"file-{index}",
            "note": f"note-{index}",
            "error-messages": [
                f"error-message-{index}-{j}"
                for j in range(message_count)
            ]
        }
        for index in range(object_count)
    ]


def test_format_error_with_duplicates():

    object_count = 10
    message_count = 3

    json_objects = get_json_objects(object_count, message_count)
    failed_object_indices = [
        index
        for index in range(object_count)
        if json_objects[index]["success"] is False
    ]

    # Check single appearance
    result = _format_json_error_messages(json_objects)
    lines = result.splitlines()
    del lines[0]
    assert_equal(len(lines), len(failed_object_indices) * 4)
    for i, failed_index in enumerate(failed_object_indices):
        # Accommodate for non-consistent message formatting, I don't
        # want to change the formatting because external tools might
        # depend on it.
        if i == 0:
            assert_equal(lines[4 * i], f"> note-{failed_index}")
        else:
            assert_equal(lines[4 * i], f">  note-{failed_index}")
        for j in range(message_count):
            assert_equal(
                lines[4 * i + 1 + j],
                f"error-message-{failed_index}-{j}")

    # Check double appearance
    result = _format_json_error_messages(json_objects + json_objects)
    lines = result.splitlines()
    del lines[0]
    assert_equal(len(lines), len(failed_object_indices) * 4)

    for i, failed_index in enumerate(failed_object_indices):
        if i == 0:
            assert_equal(lines[4 * i], f"> note-{failed_index}")
        else:
            assert_equal(lines[4 * i], f">  note-{failed_index}")
        for j in range(message_count - 1):
            assert_equal(
                lines[4 * i + 1 + j],
                f"error-message-{failed_index}-{j}")
        j = message_count - 1
        assert_equal(
            lines[4 * i + 1 + j],
            f"error-message-{failed_index}-{j} [2 times]")


def test_format_no_errors():
    json_objects = get_json_objects(1, 3)

    result = _format_json_error_messages(json_objects)
    assert_equal(result, "")

    result = _format_json_error_messages(json_objects + json_objects)
    assert_equal(result, "")


def test_command_error_rendering():
    command_error = CommandError(
        cmd="<cmd>",
        msg="<msg>",
        code=1,
        stdout="<stdout>",
        stderr="<stderr>",
        cwd="<cwd>",
        kwarg1="<kwarg1>",
        kwarg2="<kwarg2>")

    result = command_error.to_str()
    assert_equal(
        result,
        "CommandError: '<cmd>' failed with exitcode 1 under <cwd> [<msg>] "
        "[info keys: kwarg1, kwarg2] [out: '<stdout>'] [err: '<stderr>']"
    )

    result = command_error.to_str(False)
    assert_equal(
        result,
        "CommandError: '<cmd>' failed with exitcode 1 under <cwd> [<msg>] "
        "[info keys: kwarg1, kwarg2]"
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Module to help maintain a registry of versions for external modules etc
"""
import re
import sys
import os.path as op
from os import linesep

from itertools import chain
from looseversion import LooseVersion

from datalad.log import lgr
# import version helper from config to have only one implementation
# config needs this to avoid circular imports
from datalad.config import get_git_version as __get_git_version
from .exceptions import (
    CapturedException,
    CommandError,
)

__all__ = ['UnknownVersion', 'ExternalVersions', 'external_versions']


# To depict an unknown version, which can't be compared by mistake etc
class UnknownVersion:
    """For internal use
    """

    def __str__(self):
        return "UNKNOWN"

    def __cmp__(self, other):
        if other is self:
            return 0
        raise TypeError("UNKNOWN version is not comparable")


#
# Custom handlers
#
from datalad.cmd import (
    WitlessRunner,
    GitWitlessRunner,
    StdOutErrCapture,
)
from datalad.support.exceptions import (
    MissingExternalDependency,
    OutdatedExternalDependency,
)
_runner = WitlessRunner()
_git_runner = GitWitlessRunner()


def _get_annex_version():
    """Return version of available git-annex"""
    try:
        return _runner.run(
            'git annex version --raw'.split(),
            protocol=StdOutErrCapture)['stdout']
    except CommandError:
        # fall back on method that could work with older installations
        out = _runner.run(
            ['git', 'annex', 'version'],
            protocol=StdOutErrCapture)
        return out['stdout'].splitlines()[0].split(':')[1].strip()


def _get_git_version():
    """Return version of git we use (might be bundled)"""
    return __get_git_version()


def _get_system_git_version():
    """Return version of git available system-wide

    Might be different from the one we are using, which might be
    bundled with git-annex
    """
    return __get_git_version(_runner)


def _get_bundled_git_version():
    """Return version of git bundled with git-annex.
    """
    path = _git_runner._get_bundled_path()
    if path:
        out = _runner.run(
            [op.join(path, "git"), "version"],
            protocol=StdOutErrCapture)['stdout']
        # format: git version 2.22.0
        return out.split()[2]


def _get_ssh_version(exe=None):
    """Return version of ssh

    Annex prior 20170302 was using bundled version, then across all systems
    we used system one if installed, and then switched to the one defined in
    configuration, with system-wide (not default in PATH e.g. from conda)
    "forced" on Windows.  If no specific executable provided in `exe`, we will
    use the one in configuration
    """
    if exe is None:
        from datalad import cfg
        exe = cfg.obtain("datalad.ssh.executable")
    out = _runner.run(
        [exe, '-V'],
        protocol=StdOutErrCapture)
    # apparently spits out to err but I wouldn't trust it blindly
    stdout = out['stdout']
    if out['stderr'].startswith('OpenSSH'):
        stdout = out['stderr']
    match = re.match(
        "OpenSSH.*_([0-9][0-9]*)\\.([0-9][0-9]*)(p([0-9][0-9]*))?",
        stdout)
    if match:
        return "{}.{}p{}".format(
            match.groups()[0],
            match.groups()[1],
            match.groups()[3])
    raise AssertionError(f"no OpenSSH client found: {stdout}")


def _get_system_ssh_version():
    """Return version of the default on the system (in the PATH) ssh
    """
    return _get_ssh_version("ssh")


def _get_system_7z_version():
    """Return version of 7-Zip"""
    out = _runner.run(
        ['7z'],
        protocol=StdOutErrCapture)
    # reporting in variable order across platforms
    # Linux: 7-Zip [64] 16.02
    # Windows: 7-Zip 19.00 (x86)
    pieces = out['stdout'].strip().split(':', maxsplit=1)[0].strip().split()
    for p in pieces:
        # the one with the dot is the version
        if '.' in p:
            return p
    lgr.debug("Could not determine version of 7z from stdout. %s", out)


class ExternalVersions(object):
    """Helper to figure out/use versions of the externals (modules, cmdline tools, etc).

    To avoid collision between names of python modules and command line tools,
    prepend names for command line tools with `cmd:`.

    It maintains a dictionary of `distuil.version.LooseVersion`s to make
    comparisons easy. Note that even if version string conform the StrictVersion
    "standard", LooseVersion will be used.  If version can't be deduced for the
    external, `UnknownVersion()` is assigned.  If external is not present (can't
    be imported, or custom check throws exception), None is returned without
    storing it, so later call will re-evaluate fully.
    """

    UNKNOWN = UnknownVersion()

    _CUSTOM = {
        'cmd:annex': _get_annex_version,
        'cmd:git': _get_git_version,
        'cmd:bundled-git': _get_bundled_git_version,
        'cmd:system-git': _get_system_git_version,
        'cmd:ssh': _get_ssh_version,
        'cmd:system-ssh': _get_system_ssh_version,
        'cmd:7z': _get_system_7z_version,
    }
    # ad-hoc hardcoded map for relevant Python packages which do not provide
    # __version__ and are shipped by a differently named pypi package
    _PYTHON_PACKAGES = {  # Python package -> distribution package
        'github': 'pygithub',
    }
    _INTERESTING = (
        'annexremote',
        'platformdirs',
        'boto',
        'git',
        'gitdb',
        'humanize',
        'iso8601',
        'keyring',
        'keyrings.alt',
        'msgpack',
        'patool',
        'cmd:7z',
        'requests',
        'scrapy',
    )

    def __init__(self):
        self._versions = {}
        self.CUSTOM = self._CUSTOM.copy()
        self.INTERESTING = list(self._INTERESTING)  # make mutable for `add`

    @classmethod
    def _deduce_version(klass, value):
        version = None

        # see if it is something containing a version
        for attr in ('__version__', 'version'):
            if hasattr(value, attr):
                version = getattr(value, attr)
                break

        # try pkg_resources
        if version is None and hasattr(value, '__name__'):
            pkg_name = klass._PYTHON_PACKAGES.get(value.__name__, value.__name__)
            try:
                import pkg_resources
                version = pkg_resources.get_distribution(pkg_name).version
            except Exception:
                pass

        # assume that value is the version
        if version is None:
            version = value

        # do type analysis
        if isinstance(version, (tuple, list)):
            #  Generate string representation
            version = ".".join(str(x) for x in version)
        elif isinstance(version, bytes):
            version = version.decode()
        elif isinstance(version, str):
            pass
        else:
            version = None

        if version:
            return LooseVersion(version)
        else:
            return klass.UNKNOWN

    def __getitem__(self, module):
        # when ran straight in its source code -- fails to discover nipy's version.. TODO
        #if module == 'nipy':
        #    import pdb; pdb.set_trace()
        if not isinstance(module, str):
            modname = module.__name__
        else:
            modname = module
            module = None

        lgr.log(5, "Requested to provide version for %s", modname)
        # Early returns None so we do not store prev result for  them
        # and allow users to install things at run time, so later check
        # doesn't pick it up from the _versions
        if modname not in self._versions:
            version = None   # by default -- not present
            if modname in self.CUSTOM:
                try:
                    version = self.CUSTOM[modname]()
                    version = self._deduce_version(version)
                except Exception as exc:
                    lgr.debug("Failed to deduce version of %s due to %s"
                              % (modname, CapturedException(exc)))
                    return None
            else:
                if module is None:
                    if modname not in sys.modules:
                        try:
                            module = __import__(modname)
                        except ImportError:
                            lgr.debug("Module %s seems to be not present", modname)
                            return None
                        except Exception as exc:
                            lgr.warning("Failed to import module %s due to %s",
                                        modname, CapturedException(exc))
                            return None
                    else:
                        module = sys.modules[modname]
                if module:
                    version = self._deduce_version(module)
            self._versions[modname] = version

        return self._versions.get(modname, self.UNKNOWN)

    def keys(self, query=False):
        """Return names of the known modules

        Parameters
        ----------
        query: bool, optional
          If True, we will first query all CUSTOM and INTERESTING entries
          to make sure we have them known.
        """
        if query:
            [self[k] for k in chain(self.CUSTOM, self.INTERESTING)]
        return self._versions.keys()

    def __contains__(self, item):
        return bool(self[item])

    def add(self, name, func=None):
        """Add a version checker

        This method allows third-party libraries to define additional checks.
        It will not add `name` if already exists.  If `name` exists and `func`
        is different - it will override with a new `func`.  Added entries will
        be included in the output of `dumps(query=True)`.

        Parameters
        ----------
        name: str
          Name of the check (usually a name of the Python module, or an
          external command prefixed with "cmd:")
        func: callable, optional
          Function to be called to obtain version information. This should be
          defined when checking the version of something that is not a Python
          module or when this class's method for determining the version of a
          Python module isn't sufficient.
        """
        if func:
            func_existing = self.CUSTOM.get(name, None)
            was_known = False
            if func_existing and func_existing is not func:
                lgr.debug(
                    "Adding a new custom version checker %s for %s, "
                    "old one: %s", func, name, func_existing)
                was_known = name in self._versions
            self.CUSTOM[name] = func
            if was_known:
                # pop and query it again right away to possibly replace with a new value
                self._versions.pop(name)
                _ = self[name]
        elif name not in self.INTERESTING:
            self.INTERESTING.append(name)

    @property
    def versions(self):
        """Return dictionary (copy) of versions"""
        return self._versions.copy()

    def dumps(self, indent=None, preamble="Versions:", query=False):
        """Return listing of versions as a string

        Parameters
        ----------
        indent: bool or str, optional
          If set would instruct on how to indent entries (if just True, ' '
          is used). Otherwise returned in a single line
        preamble: str, optional
          What preamble to the listing to use
        query : bool, optional
          To query for versions of all "registered" custom externals, so to
          get those which weren't queried for yet
        """
        if indent and (indent is True):
            indent = ' '
        items = ["%s=%s" % (k, self._versions[k]) for k in sorted(self.keys(query=query))]
        out = "%s" % preamble if preamble else ''
        if indent is not None:
            if preamble:
                preamble += linesep
            indent = ' ' if indent is True else str(indent)
            out += (linesep + indent).join(items) + linesep
        else:
            out += " " + ' '.join(items)
        return out

    def check(self, name, min_version=None, msg=""):
        """Check if an external (optionally of specified min version) present

        Parameters
        ----------
        name: str
          Name of the external (typically a Python module)
        min_version: str or version, optional
          Minimal version to satisfy
        msg: str, optional
          An additional message to include into the exception message

        Raises
        ------
        MissingExternalDependency
          if the external is completely missing
        OutdatedExternalDependency
          if the external is present but does not satisfy the min_version
        """
        ver_present = self[name]
        if ver_present is None:
            raise MissingExternalDependency(
                name, ver=min_version, msg=msg)
        elif min_version and ver_present < min_version:
            raise OutdatedExternalDependency(
                name, ver=min_version, ver_present=ver_present, msg=msg)


external_versions = ExternalVersions()
